import numpy as np
import h5py


class CloneNeurons:

    def __init__(self, source_network_file, dest_network_file, neuron_type):

        """ This code copies neurons of neuron_type from one network to another.
            The number of neurons must match in the two networks. """
        
        self.source_file = h5py.File(source_network_file,"r")
        self.dest_file = h5py.File(dest_network_file, "a")
        self.neuron_type = neuron_type

    def close(self):
        self.source_file.close()
        self.dest_file.close()
        
    def get_neuron_id(self, hdf5_file, neuron_type):

        neuron_id = [ x for x,y in zip(hdf5_file["network/neurons/neuron_id"],
                                       hdf5_file["network/neurons/name"])
                      if y.decode().split("_")[0] == neuron_type]
        
        return np.array(neuron_id)

    def clone(self, source_id, dest_id):

        print(f"Cloning source_id={source_id} --> dest_id={dest_id}")
        
        for var_name in self.source_file["network/neurons"].keys():

            if var_name == "neuron_id":
                # Keep original dest neuron ID
                continue

            if len(self.source_file["network/neurons"][var_name].shape) == 1:
                self.dest_file["network/neurons"][var_name][dest_id] = \
                    self.source_file["network/neurons"][var_name][source_id]
            else:
                self.dest_file["network/neurons"][var_name][dest_id, :] = \
                    self.source_file["network/neurons"][var_name][source_id, :]

    def clone_neurons(self, neuron_type=None):

        if neuron_type is None:
            neuron_type = self.neuron_type

        source_id = self.get_neuron_id(self.source_file, neuron_type)
        dest_id = self.get_neuron_id(self.dest_file, neuron_type)

        assert len(source_id) == len(dest_id), \
            f"There must be equal number of {neuron_type} in the two networks"
        
        for s_id, d_id in zip(source_id, dest_id):
            self.clone(source_id = s_id, dest_id = d_id)

            
if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("source_file")
    parser.add_argument("dest_file")
    parser.add_argument("neuron_type")

    args = parser.parse_args()

    cn = CloneNeurons(source_network_file=args.source_file,
                      dest_network_file=args.dest_file,
                      neuron_type=args.neuron_type)

    cn.clone_neurons()
    cn.close()



import sys
import os.path

import neuron
import numpy as np

from snudda.neurons.neuron_prototype import NeuronPrototype
from snudda.simulate.nrn_simulator_parallel import NrnSimulatorParallel

import bluepyopt.ephys as ephys
from snudda.neurons.neuron_model_extended import NeuronModel
from snudda.neurons.neuron_morphology import NeuronMorphology


# Plot all sections
# [neuron.h.psection(x) for x in neuron.h.allsec()]

##############################################################################


class RunSynapseRun(object):

    def __init__(self,
                 stim_times,
                 synapse_density,
                 num_synapses,
                 neuron_parameter_key,
                 neuron_morphology_key,
                 neuron_path=None,
                 neuron_morphology=None,
                 neuron_mechanisms=None,
                 neuron_parameters=None,
                 # neuron_modulation=None,
                 synapse_section_id=None,  # if given, nSynapses is ignored
                 synapse_section_x=None,  # same # of elements as synapseSectionID
                 # neuron_parameter_id=0,  # Which param set in parameter file to use
                 # neuron_morphology_id=0,
                 # neuron_modulation_id=0,
                 holding_voltage=-70e-3,
                 holding_current=None,
                 synapse_type='glut',
                 params={},
                 time=2,
                 random_seed=None,
                 rng=None,
                 log_file=None,
                 verbose=True):

        self.log_file = log_file  # File pointer
        self.verbose = verbose

        if rng is None:
            self.rng = np.random.default_rng(random_seed)
        else:
            self.rng = rng

        self.write_log(f"Holding voltage: {holding_voltage} V")
        self.write_log(f"Stim times: {stim_times} s")
        self.write_log(f"Synapse type: {synapse_type}")

        self.time = time
        self.synapses = []
        self.i_clamp = None
        self.nc_syn = None

        self.i_save = None
        self.t_save = None
        self.v_save = None

        self.stim_vector = None
        self.vec_stim = None
        self.synapse_section_id = None
        self.synapse_section_x = None

        if neuron_path is None and neuron_parameters is not None:
            neuron_path = os.path.dirname(neuron_parameters)

        self.neuron_path = neuron_path

        if neuron_parameters is None and self.neuron_path is not None:
            neuron_parameters = os.path.join(self.neuron_path, "parameters.json")

        if neuron_morphology is None and self.neuron_path is not None:
            neuron_morphology = os.path.join(self.neuron_path, "morphology")

        if neuron_mechanisms is None and self.neuron_path is not None:
            neuron_mechanisms = os.path.join(self.neuron_path, "mechanisms.json")

        # if neuron_modulation is None and self.neuron_path is not None:
        #     neuron_modulation = os.path.join(self.neuron_path, "modulation.json")

        # Done in NrnSimulatorParallel
        # neuron.h.load_file('stdrun.hoc')

        self.sim = NrnSimulatorParallel(cvode_active=False)

        # Should we use weak reference for garbage collection? (weakref package)

        # We load the neuron morphology object also, used to place synapses
        self.write_log(f"Using morphology: {neuron_morphology}")
        neuron_prototype = NeuronPrototype(neuron_path=neuron_path,
                                           neuron_name="OptimisationNeuron")
        self.morphology = neuron_prototype.clone(parameter_key=neuron_parameter_key,
                                                 morphology_key=neuron_morphology_key)
        # self.morphology = NeuronMorphology(swc_filename=neuron_morphology)

        # We need to setup the Neuron model
        self.neuron = NeuronModel(param_file=neuron_parameters,
                                  morph_path=neuron_morphology,
                                  mech_file=neuron_mechanisms,
                                  cell_name="OptimisationNeuron",
                                  # modulation_file=neuron_modulation,
                                  parameter_key=neuron_parameter_key,
                                  morphology_key=neuron_morphology_key)

        self.neuron.instantiate(sim=self.sim)
        self.set_resting_voltage(holding_voltage * 1e3)

        neuron.h.celsius = 35

        # gnabar_hh: The maximum specific sodium channel conductance [Default value = 0.120 S/cm2]
        # gkbar_hh: The maximum specific potassium channel conductance [Default value = 0.036 S/cm2]
        # gl_hh: The maximum specific leakage conductance [Default value = 0.0003 S/cm2]
        # ena: The reversal potential for the sodium channel [Default value = 50 mV]
        # ek: The reversal potential for the potassium channel [Default value = -77 mV]
        # el_hh: The reversal potential for the leakage channel [Default value = -54.3 mV]

        # We need to set the params also
        self.params = params
        self.default_cond = 5e-7

        # This returns (section,sectionX) so we can reuse it if needed
        self.synapse_positions = self.add_synapse_density(synapse_type=synapse_type,
                                                          synapse_density=synapse_density,
                                                          num_synapses=num_synapses,
                                                          section_id=synapse_section_id,
                                                          section_x=synapse_section_x)

        self.stim_times = np.array(stim_times)

        # Assumes input in seconds (SI units)
        self.connect_input_to_synapses(self.stim_times)

        self.soma_record()
        self.synapse_current_record()

        self.holding_current = self.update_holding_current(holding_voltage=holding_voltage,
                                                           holding_current=holding_current)

        self.write_log("RunSynapseRun: Init done.")

        # import pdb
        # pdb.set_trace()

    ############################################################################

    def __del__(self):

        # This should not be needed but...
        self.neuron = None
        self.morphology = None
        self.i_clamp = None
        self.v_clamp = None
        self.v_save = None
        self.t_save = None
        self.i_save = None
        self.nc_syn = None

        self.vec_stim = None
        self.stim_vector = None
        self.little_synapse = None

    ############################################################################

    def update_holding_current(self, holding_voltage=None, holding_current=None):

        if holding_voltage is None:
            holding_voltage = self.holding_voltage
        else:
            self.holding_voltage = holding_voltage

        if holding_current is not None:
            if self.i_clamp is None:
                self.i_clamp = neuron.h.IClamp(self.neuron.icell.soma[0](0.5))

            # Update current on iClamp
            self.i_clamp.amp = holding_current * 1e9  # Convert SI -> nA for NEURON
            self.i_clamp.dur = 2 * self.time * 1e3

            self.set_resting_voltage(self.holding_voltage * 1e3)
            self.write_log(f"Set holding current {holding_current}A and holding voltage {holding_voltage}V,"
                           f" until {self.i_clamp.dur} ms")
            return holding_current

        self.write_log("Updating holding current, might take a bit of time")

        # Disable old iClamp temporarily
        if self.i_clamp is not None:
            self.i_clamp.amp = 0

        # Setup a temporary VClamp
        self.v_clamp = neuron.h.SEClamp(self.neuron.icell.soma[0](0.5))
        self.v_clamp.rs = 1e-9
        self.v_clamp.amp1 = holding_voltage * 1e3
        self.v_clamp.dur1 = self.time * 2 * 1e3
        # self.writeLog("VClamp duration: " + str(self.VClamp.dur1))

        neuron.h.finitialize(self.holding_voltage * 1e3)
        # !!! There is a WEIRD neuron bug, that if this tstop here is
        # different from duration of simulation, then the *SECOND* time
        # a model is initialised we get the length of tSave set by this
        # value, and not by the tStop of that simulation --- go figure!
        self.set_resting_voltage(self.holding_voltage * 1e3)

        neuron.h.tstop = self.time * 1e3  # Must set tstop
        neuron.h.run()

        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(self.t_save, self.v_save)
            plt.title("Holding voltage should be " + str(self.holding_voltage * 1e3) + "mV")
            plt.xlabel("time (ms)")
            plt.ylabel("volt (mV)")
            plt.ion()
            plt.show()

            import pdb
            pdb.set_trace()

        cur = float(self.v_clamp.i)

        # Remove VClamp
        self.v_clamp = None

        if self.i_clamp is None:
            self.i_clamp = neuron.h.IClamp(self.neuron.icell.soma[0](0.5))

        # Update current on iClamp
        self.i_clamp.amp = cur
        self.i_clamp.dur = 2 * self.time * 1e3

        self.set_resting_voltage(self.holding_voltage * 1e3)

        self.write_log(f"Holding voltage {self.holding_voltage * 1e3} mV, IClamp amp = {cur} nA")

        return cur * 1e-9  # Convert to SI units

    ############################################################################

    def set_stim_times(self, stim_times):

        if len(stim_times) != len(self.stim_times) or (stim_times != self.stim_times).any():
            print(f"Setting stim times to {stim_times} s")
            self.write_log(f"Setting stim times to {stim_times} s")
            self.stim_vector = neuron.h.Vector(stim_times * 1e3)
            self.stim_times = stim_times * 1e3

    ############################################################################

    def add_synapse_density(self, synapse_type,
                            synapse_density,
                            num_synapses=None,
                            plot_flag=False,
                            section_id=None,
                            section_x=None):

        if plot_flag:

            assert section_id is None and section_x is None, \
                "Can not plot if sectionID and sectionX are given"

            input_coords, section_id, section_x, density_function, dist_syn_soma = \
                self.morphology.dendrite_input_locations(synapse_density_str=synapse_density,
                                                         num_locations=num_synapses,
                                                         rng=self.rng,
                                                         return_density=True)

            self.synapse_locations = input_coords
            dist_from_soma = self.morphology.dend[:, 4]

            # plot density function
            plt.figure()
            plt.plot(dist_from_soma * 1e6, density_function, 'o')
            plt.xlabel('distance from soma $(\mu m)$')
            plt.title('density function')
            plt.ion()
            plt.show()

            # plot histogram - distance synapses from soma
            plt.figure()
            plt.hist(dist_syn_soma * 1e6, edgecolor='gray', bins=dist_syn_soma.size)
            plt.xlabel('distance from soma $(\mu m)$')
            plt.ylabel('frequency')
            plt.title('synaptic distance from soma')
            plt.ion()
            plt.show()

        elif section_id is None or section_x is None:

            input_coords, section_id, section_x, dist_syn_soma = \
                self.morphology.dendrite_input_locations(synapse_density_str=synapse_density,
                                                         num_locations=num_synapses,
                                                         rng=self.rng)

            self.synapse_locations = input_coords

        else:
            self.synapse_locations = "Unknown, you specified sectionX and sectionID, " \
                                     + "but not synapse xyz coordinates."

        # This is so we can find out where the synapses were placed
        self.synapse_section_id = section_id
        self.synapse_section_x = section_x

        sections = self.neuron.map_id_to_compartment(section_id)

        for s, sx in zip(sections, section_x):
            self.add_synapse(synapse_type, s, sx, self.params)

        # Return synapse position if we want to reuse them elsewhere
        return sections, section_x

    ############################################################################

    def add_synapse(self, synapse_type, section, section_x, params):

        section_x = np.maximum(section_x, 1e-6)  # Cant be 0 or 1

        try:
            if synapse_type.lower() == 'glut':
                syn = neuron.h.tmGlut_double(section(section_x))
            elif synapse_type.lower() == "gaba":
                syn = neuron.h.tmGabaA(section(section_x))
            else:
                assert False, f"Unknown synapse type: {synapse_type}"

            self.synapses.append(syn)

        except:
            import traceback
            tstr = traceback.format_exc()
            self.write_log(tstr)

            self.write_log("Did you remember to run nrnivmodl first, to generate channels mod files?")
            sys.exit(-1)

        for p in params:

            # Conductance is set separately, it is a property of netcon
            if p == "cond":
                self.default_cond = params["cond"]
                continue

            val = self.si_to_natural_units(p, params[p])

            setattr(syn, p, val)
            # self.write_log(f"Setting parameters: {p} = {val} (neuron natural units)")

    ############################################################################

    def connect_input_to_synapses(self, stim_times):

        self.write_log(f"Stimulation times (s): {stim_times}")
        self.nc_syn = []

        self.stim_vector = neuron.h.Vector(stim_times * 1e3)
        self.vec_stim = neuron.h.VecStim()
        self.vec_stim.play(self.stim_vector)

        for syn in self.synapses:
            ncs = neuron.h.NetCon(self.vec_stim, syn)
            ncs.delay = 0
            ncs.threshold = 0
            self.nc_syn.append(ncs)

    ############################################################################

    def soma_record(self):

        self.t_save = neuron.h.Vector()
        self.t_save.record(neuron.h._ref_t)

        self.v_save = neuron.h.Vector()
        self.v_save.record(self.neuron.icell.soma[0](0.5)._ref_v)

    ############################################################################

    def synapse_current_record(self):

        self.i_save = []

        for syn in self.synapses:
            i_rec = neuron.h.Vector()
            i_rec.record(syn._ref_i)
            self.i_save.append(i_rec)

    ############################################################################

    def run(self, tau, tau_r, tau_f, u, cond=None, time=None):

        assert False, "This is the old run method"

        if time is None:
            time = self.time
        else:
            self.time = time

        if cond is None:
            cond = self.default_cond

        # print(vars())

        # print("Running: tau: %.3g, tauR: %.3g, tauF: %.3g, U: %.3g, cond: %.3g\n" \
        #      % (tau,tauR,tauF,U,cond))

        # Convert from SI units to natural units that Neuron uses
        for ncs in self.nc_syn:
            ncs.weight[0] = 1 * cond * 1e6

        for syn in self.synapses:
            syn.tau = tau * 1e3
            syn.tau_r = tau_r * 1e3
            syn.tau_f = tau_f * 1e3
            syn.u = u

        # print(self.littleSynapse.tau)

        # print("Initialise voltage to " + str(self.holdingVoltage*1e3) + " mV")
        neuron.h.finitialize(self.holding_voltage * 1e3)  # OLD : -70
        neuron.h.tstop = time * 1e3

        neuron.h.run()

        # self.tSave.resize()
        # self.vSave.resize()
        # self.iSave.resize()

        if np.array(self.t_save).shape != np.array(self.v_save).shape:
            self.write_log("THIS IS WRONG, should be same shape!!")
            self.write_log(f"size t = {np.array(self.t_save).shape}")
            self.write_log(f"size v = {np.array(self.v_save).shape}")
            import pdb
            pdb.set_trace()

        # print("Last V = " + str(self.vSave[-1]*1e-3))

        # Convert back to SI units
        return (np.array(self.t_save) * 1e-3,
                np.array(self.v_save) * 1e-3,
                np.array(self.i_save) * 1e-9)

    ############################################################################

    def set_resting_voltage(self, rest_volt):

        self.write_log("Setting resting voltage to %.3f mV" % rest_volt)

        soma = [x for x in self.neuron.icell.soma]
        axon = [x for x in self.neuron.icell.axon]
        dend = [x for x in self.neuron.icell.dend]

        cell = soma + axon + dend

        for sec in cell:
            for seg in sec.allseg():
                seg.v = rest_volt

    ############################################################################

    # I wish Neuron would use natural units...

    def si_to_natural_units(self, var_name, value):

        conv_factor = {"U": 1.0,
                       "tauR": 1e3,
                       "tauF": 1e3,
                       "cond": 1e6,
                       "tau": 1e3,
                       "nmda_ratio": 1.0,

                       "tau1_ampa": 1.0,  # Ilaria's file has ms already
                       "tau2_ampa": 1.0,  # Ilaria's file has ms already
                       "tau3_ampa": 1.0,  # Ilaria's file has ms already
                       "tau1_nmda": 1.0,  # Ilaria's file has ms already
                       "tau2_nmda": 1.0,  # Ilaria's file has ms already
                       "tau3_nmda": 1.0,  # Ilaria's file has ms already
                       "tpeak_ampa": 1.0,  # Ilaria's file has ms already
                       "tpeak_nmda": 1.0,  # Ilaria's file has ms already :/
                       "ratio_nmda": 1.0,

                       "I2_ampa": 1.0,  # Ilaria's file has ms already
                       "I3_ampa": 1.0,  # Ilaria's file has ms already
                       "I2_nmda": 1.0,  # Ilaria's file has ms already
                       "I3_nmda": 1.0,  # Ilaria's file has ms already
                       "factor_ampa": 1.0,  # Ilaria's file has ms already
                       "factor_nmda": 1.0  # Ilaria's file has ms already
                       }

        if var_name not in conv_factor:
            self.write_log("Missing conversion fractor for " + str(var_name) \
                           + ". Please update SItoNaturalUnits function.")
            self.write_log("convFactor = " + str(conv_factor))
            import pdb
            pdb.set_trace()

        try:
            return value * conv_factor[var_name]
        except:
            import traceback
            tstr = traceback.format_exc()
            self.write_log(tstr)
            import pdb
            pdb.set_trace()

    ############################################################################
    def plot(self):

        ax = self.morphology.plot_neuron(axon_colour='red', dend_colour='blue', soma_colour='green')

        ax.scatter(self.synapse_locations[:, 0],
                   self.synapse_locations[:, 1],
                   self.synapse_locations[:, 2],
                   color='black',
                   marker='o',
                   s=20)
        plt.ion()

        ############################################################################

    # pars = { "tau1" : 0.25e-3 }
    # The code converts to natural units internally, if your parameter is missing
    # then update SItoNaturalUnits

    # OBS, soma parameters are ignored by run2 (they should be set during setup)

    def run2(self, pars, time=None, cond=1e-8):

        self.write_log(f"Running with pars: {pars}")

        if time is None:
            time = self.time
        else:
            self.time = time

        for p in pars:

            if p == "cond":
                cond = self.si_to_natural_units(p, pars[p])

            else:
                v = self.si_to_natural_units(p, pars[p])
                for s in self.synapses:
                    setattr(s, p, v)

        neuron.h.finitialize(self.holding_voltage * 1e3)
        for ncs in self.nc_syn:
            ncs.weight[0] = cond

        self.set_resting_voltage(self.holding_voltage * 1e3)

        neuron.h.v_init = self.holding_voltage * 1e3
        neuron.h.tstop = time * 1e3
        self.write_log("About to start NEURON... stay safe")
        neuron.h.run()
        self.write_log("NEURON actually completed?!")

        # Convert results back to SI units
        return (np.array(self.t_save) * 1e-3,
                np.array(self.v_save) * 1e-3,
                np.array(self.i_save) * 1e-9)

    ############################################################################

    def write_log(self, text, flush=True):  # Change flush to False in future, debug
        if self.log_file is not None:
            self.log_file.write(text + "\n")

            if self.verbose:
                print(text)

            if flush:
                self.log_file.flush()
        else:
            if self.verbose:
                print(text)

            ##############################################################################


if __name__ == "__main__":

    import matplotlib.pyplot as plt
    import time

    start = time.time()

    stim_times = np.array([0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 1.0]) + 0.3

    # synapseDensity="0.05/(1+np.exp(-(d-30e-6)/5e-6))"
    # synapseDensity="0.05/(1+np.exp(-(d-150e-6)/5e-6))"

    synapse_density = "np.exp(-((d-50e-6)/15e-6)**2)"
    holding_voltage = -79.5e-3

    # !!! OBS changed number of synapses, before was 20 -- cond 1nS
    tau = 10e-3
    tau_r = 200e-3
    tauF = 900e-3
    U = 0.5

    tau = 0.1
    tau_r = 0.1
    tauF = 0.1
    U = 0.1
    cond = 1e-9

    rlsr = RunSynapseRun(stim_times=stim_times,
                         synapse_type="glut",
                         neuron_morphology="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/WT-1215MSN03-cor-rep-ax2.swc",
                         neuron_parameters="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/parameters.json",
                         neuron_mechanisms="../data/cellspecs/dspn/str-dspn-e150917_c9_d1-mWT-1215MSN03-v20190521/mechanisms.json",
                         synapse_density=synapse_density,
                         num_synapses=10,
                         neuron_parameter_key=0,
                         # neuron_modulation_key=0,
                         holding_voltage=holding_voltage)

    # I would like to plot the morphology
    # ax = rlsr.plot()

    # import pdb
    # pdb.set_trace()

    params = {"U": 0.5,
              "tauR": 0.6,
              "tauF": 2.0,  # 1.1
              "tau": 0.08,
              "cond": 1e-9,
              "nmda_ratio": 0.5}

    tS, vS, iS = rlsr.run2(pars=params)
    plt.figure()
    plt.plot(tS, vS)
    plt.xlabel("Time (s)")
    plt.ylabel("Voltage (V)")

    if True:
        expData = 'DATA/YvonneJohansson2019/M1LH_Analysis_191001.h5'
        expCellID = 131

        import LoadExpData

        led = LoadExpData.LoadExpData(expData)
        cellType = led.get_cell_type(expCellID)
        plt.title("ExpData is cellType : " + str(cellType))
        expV, expT = led.getData("GBZ_CC_H20", expCellID)
        plt.plot(expT, expV, 'r-')

    if False:
        plt.figure()

        for iiS in iS:
            plt.plot(tS, iiS)

        plt.xlabel("Time (s)")
        plt.ylabel("Current (A)")

    plt.ion()
    plt.show()

    end = time.time()
    print(end - start)

    import pdb

    pdb.set_trace()



import sys
from collections import OrderedDict
import os
import heapq
import numpy as np
import json
import datetime
import copy

from snudda.utils.numpy_encoder import  NumpyEncoder


class ParameterBookkeeper:

    def __init__(self, n_max=10, old_book=None, old_book_file=None):

        self.n_max = n_max
        self.old_iter = 0

        if old_book:
            self.book = copy.deepcopy(old_book)
            heapq.heapify(self.book)
        else:
            self.book = []

        if old_book_file and os.path.exists(old_book_file):
            self.load(old_book_file)

    def add_parameters(self, parameter_set, section_id, section_x, error, dt=None, volt=None):

        now = datetime.datetime.now().timestamp()

        data = OrderedDict()
        data["parameters"] = parameter_set
        data["section_id"] = section_id
        data["section_x"] = section_x
        data["error"] = error

        assert error >= 0, f"Error should be positive: error given {error}"

        if dt is not None:
            data["dt"] = dt
        if volt is not None:
            data["volt"] = volt

        if len(self.book) >= self.n_max:
            worst_error = self.book[0][0]
            if error > np.abs(worst_error):
                # The new parameter is worse than the current worst one, don't add since we are already full
                return

        try:
            heapq.heappush(self.book, (-error, now, data))
        except:
            import traceback
            tstr = traceback.format_exc()
            print(tstr)
            sys.exit(-1)

        book_len = len(self.book)
        for i in range(self.n_max, book_len):
            heapq.heappop(self.book)

    def merge(self, *other_books):

        # We need to heapify the returned list before we use it as a heap again
        self.book = list(heapq.merge(self.book, *other_books))
        heapq.heapify(self.book)

        # We need to remove surplus elements
        book_len = len(self.book)
        for i in range(self.n_max, book_len):
            heapq.heappop(self.book)

    def get_dictionary(self):
        data_dict = OrderedDict()
        data_list = []

        book_copy = list(self.book)

        for i in range(0, len(self.book)):
            _, _, data = heapq.heappop(book_copy)
            data_list.insert(0, data)

        for idx, data in enumerate(data_list):
            data_dict[idx] = data

        return data_dict

    def get_best_parameterset(self):
        if len(self.book) == 0:
            return None
        else:
            sorted_data = sorted(self.book, key=lambda x: -x[0])
            best_data = sorted_data[0]
            return best_data[2]["parameters"]  # First is error

    def get_best_dataset(self):
        if len(self.book) == 0:
            return None
        else:
            sorted_data = sorted(self.book, key=lambda x: -x[0])
            best_data = sorted_data[0]
            return best_data[2]  # First is error, second datetime (tiebreaker)

    def clear(self):
        self.book = []

    def load(self, file_name, clear=False):

        if clear:
            self.clear()

        with open(file_name, "r") as f:
            data = json.load(f)

        if "meta" in data:
            if "iter" in data["meta"]:
                self.old_iter = data["meta"]["iter"]

        for d in data.values():

            if "parameters" not in d:
                continue

            self.add_parameters(parameter_set=np.array(d["parameters"]),
                                section_id=np.array(d["section_id"]).astype(int),
                                section_x=np.array(d["section_x"]),
                                error=d["error"],
                                # dt=d["dt"],
                                # volt=np.array(d["volt"])
                                )

    def save(self, file_name):
        print(f"Writing parameter data to {file_name}")
        data_dict = self.get_dictionary()
        data_dict["meta"] = {"iter" : self.old_iter}

        with open(file_name, "w") as f:
            json.dump(data_dict, f, indent=4, cls=NumpyEncoder)

    def get_iter(self):
        return self.old_iter

    def set_iter(self, iter):
        self.old_iter = iter

    def check_integrity(self):

        if len(self.book) <= 1:
            return

        param_len = np.array([len(d[2]["parameters"]) for d in self.book])
        assert (param_len == param_len[0]).all(), "Parameters should all be same length"

        section_id = [d[2]["section_id"] for d in self.book]
        section_x = [d[2]["section_x"] for d in self.book]

        assert np.array([s == section_id[0] for s in section_id]).all(), \
            "section_id should match for all parametersets"

        assert np.array([s == section_x[0] for s in section_x]).all(), \
            "section_x should match for all parametersets"








import plotly.graph_objects as go
from plotly.subplots import make_subplots

import numpy as np
from snudda.utils import SnuddaLoadSimulation
from snudda.utils.load import SnuddaLoad

# To get clean plots:
#
# import plotly_plots.io as pio
# pio.templates.default = "simple_white"
#


class PlotTrace:

    def __init__(self, snudda_load_simulation=None, network_simulation_file=None):

        self.snudda_load_simulation = None
        self.colour_map = dict()
        self.traces = dict()
        self.default_colour = None
        self.default_width = 1

        if snudda_load_simulation is not None:
            self.snudda_load_simulation = snudda_load_simulation

        if network_simulation_file is not None:
            self.load_data(network_simulation_file=network_simulation_file)

    def load_data(self, network_simulation_file, do_depol_block_test=False):
        self.snudda_load_simulation = SnuddaLoadSimulation(network_simulation_output_file=network_simulation_file,
                                                           quiet_load=True, do_test=do_depol_block_test)

    def define_colour_by_neuron_id(self, colour_by_neuron_id):
        self.colour_map |= colour_by_neuron_id

    def define_colour_by_neuron_type(self, colour_by_neuron_type):
        neuron_names = [SnuddaLoad.to_str(x) for x
                        in self.snudda_load_simulation.network_simulation_file["meta_data"]["name"]]
        neuron_types = [x.split("_")[0] for x in neuron_names]

        for neuron_id, nt in enumerate(neuron_types):
            if nt in colour_by_neuron_type:
                self.colour_map[neuron_id] = colour_by_neuron_type[nt]

    def define_colour_by_neuron_name(self, colour_by_name):
        neuron_names = self.snudda_load_simulation.get_neuron_name()
        for neuron_id, name in enumerate(neuron_names):
            if name in colour_by_name:
                self.colour_map[neuron_id] = colour_by_name[name]

    def clear_colour(self):
        self.colour_map = dict()

    def get_instant_frequency(self, neuron_id):
        # neuron_id is assumed to be scalar int here
        spike_times = self.snudda_load_simulation.get_spikes(neuron_id=neuron_id).flatten()

        isi = 1.0 / np.diff(spike_times)
        isi_time = spike_times[1:]

        return isi_time, isi

    def get_voltage_trace(self, neuron_id):

        volt = self.snudda_load_simulation.get_voltage(neuron_id=neuron_id)
        time = self.snudda_load_simulation.get_time()

        return time, volt[:, 0]

    def create_trace(self, neuron_id):

        time, volt = self.get_voltage_trace(neuron_id=neuron_id)
        isi_time, isi = self.get_instant_frequency(neuron_id=neuron_id)

        colour = self.colour_map.get(neuron_id, self.default_colour)
        neuron_name = self.snudda_load_simulation.get_neuron_name(neuron_id=[neuron_id])[0]
        label = f"{neuron_name} ({neuron_id})"

        isi_sct = go.Scatter(x=isi_time.T, y=isi.T, mode="lines",
                             line={"color": colour, "width": self.default_width},
                             name=label, legendgroup=label, showlegend=False)

        volt_sct = go.Scatter(x=time.T, y=volt.T, mode="lines",
                              line={"color": colour, "width": self.default_width},
                              name=label, legendgroup=label)

        if neuron_id not in self.traces:
            self.traces[neuron_id] = dict()

        self.traces[neuron_id]["isi"] = isi_sct
        self.traces[neuron_id]["volt"] = volt_sct

    def create_traces(self, neuron_id_list=None):
        if neuron_id_list is None:
            neuron_id_list = [int(x) for x
                              in self.snudda_load_simulation.network_simulation_file["neurons"].keys()]

        for neuron_id in neuron_id_list:
            self.create_trace(neuron_id=neuron_id)

    def make_figure(self, show=False):

        fig = make_subplots(cols=1, rows=2, shared_xaxes=True)

        for neuron_id, plots in self.traces.items():
            if neuron_id not in self.colour_map:
                raise KeyError(f"No colour defined for {neuron_id = } or its neuron type")
            fig.add_trace(plots["volt"], row=1, col=1)
            fig.add_trace(plots["isi"], row=2, col=1)

        fig.update_layout(xaxis={"title": "Time (s)"},
                          yaxis1={"title": "Volt (V)"},
                          yaxis2={"title": "Frequency (Hz)"})

        if show:
            fig.show()

        return fig

    def plot_traces(self, neuron_id=None, show=False):
        if isinstance(neuron_id, (int, np.integer)):
            neuron_id = [neuron_id]

        self.create_traces(neuron_id_list=neuron_id)
        fig = self.make_figure(show=show)
        return fig

    # To save: fig.write_image("yourfigname.png")


# TODO: 2026-02-26 -- Dear future humans.
#      -- We have copied optimise_synapse_full.py (but not all of it)
#      -- we need to write code to run on self.pc.id() == 0 that sets up the optimisation
#          and performs all the goodie stuff
#
#      We got to 'get_refined_parameters' function, probably not needed?
#      We are going to use mpirun and neuron parallel, to run multiple instances at the same time.

# We want to run synapse optimisation in parallel
# 1. Start using mpirun (to get multiple instances)
# 2. Worker id 0 (master) sets up optimisation, bounds etc.
# 3. Setup models (with default parameters, will be overwritten each iteration)
# 4. Start optimisation loop on 0.
# 5. Get the paramater sets to investigate
# 6. Distribute parameters to all the workers, set them to the synapses
# 7. Simulate the simulation
# 8. Gather the results master node
# 9. Update the opt object with results
# 10. Repeat 5-10

import os
import sys
import shutil
import timeit

import numpy as np
import json
import copy
import time

import scipy

from mpi4py import MPI  # This must be imported before neuron, to run parallel
import neuron
from neuron import h  # , gui
from snudda.utils.snudda_path import snudda_parse_path, get_snudda_data
from snudda.synaptic_fitting.parameter_bookkeeper import ParameterBookkeeper

from run_synapse_run import RunSynapseRun


class SynapseOptimiser:

    def __init__(self, data_file,
                 entropy=1023456734529028340264793840,
                 synapse_type="glut",
                 load_parameters=True,
                 snudda_data=None,
                 neuron_set_file="neuron_set.json",
                 synapse_parameter_file=None,
                 verbose=True):

        self.log_file = None
        self.verbose = verbose
        self.rng = None


        self.data_file = data_file
        self.neuron_set_file = neuron_set_file
        self.seed = None
        self.entropy=entropy

        self.snudda_data = get_snudda_data(snudda_data=snudda_data)

        self.rsr_synapse_model = None
        self.synapse_type = synapse_type
        self.data = None
        self.volt = None
        self.time = None
        self.sample_freq = None
        self.trace_holding_voltage = None
        self.stim_time = None
        self.cell_type = None
        self.neuron_set_file = None

        self.synapse_parameter_data = None
        self.synapse_section_id = None
        self.synapse_section_x = None

        self.cell_properties = None

        self.load_trace_data()

        # if load_parameters:
        #     self.load_parameter_data()

        self.synapse_parameter_file = synapse_parameter_file

        if synapse_parameter_file:
            with open(synapse_parameter_file, 'r') as f:
                self.write_log(f"Reading synapse parameters from {synapse_parameter_file}")
                self.synapse_parameters = json.load(f)["data"]
        else:
            self.synapse_parameters = {}

        self.parameter_data_file_name = f"{self.data_file}-parameters-full.json"



        self.pc = h.ParallelContext()
        self.n_workers = self.pc.nhost()

        print(f"I am {self.pc.id()}/{self.pc.nhost()}")
        if self.pc.id() == 0:
            rng = np.random.default_rng(seed=0)
            param_list = [rng.random(size=(3, 1)) for x in range(self.pc.nhost())]
            print(f"Before sending: {param_list = }")
        else:
            param_list = []

        self.param = self.pc.py_scatter(param_list)

        self.pc.barrier()
        print(f"Done {self.pc.id()} {self.param.T = }")

    def setup_rng(self):

        if self.rng is not None:
            print(f"setup_rng: rng already setup, skipping.")
            return

        seeds = []

        if self.pc.id() == 0:
            # Setup and distribute random seeds to all workers
            seed_sequence = np.random.SeedSequence(entropy=self.entropy)
            seeds=seed_sequence.generate_state(self.n_workers)

        self.seed = self.pc.py_scatter(seeds)
        self.rng = np.random.default_rng(seed=self.seed)

        self.pc.barrier()


    def prepare_models(self):

        if self.pc.id() == 0:
            # Setup the model on master node, this sets self.synapse_section_id (and _x)

            self.synapse_model = self.setup_model(synapse_density_override=None,
                                                  n_synapses_override=None,
                                                  synapse_position_override=None)




        self.pc.barrier()
        # Distribute section id, section x that was picked by master, and any other needed parameters
        self.synapse_section_id, self.synapse_section_x \
            = self.pc.broadcast((self.synapse_section_id, self.synapse_section_x))

        if self.pc.id() != 0:
            # Setup models on all other nodes (but not master)
            synapse_position_override = (self.synapse_section_id, self.synapse_section_x)

            self.synapse_model = self.setup_model(synapse_position_override=synapse_position_override)

        self.pc.barrier()

    def run_models(self, model_parameter_list):

        model_parameters = self.pc.py_scatter(model_parameter_list)



        # we need model parameters, and position of synapses (section_id, section_x)

        self.run_model()

        # TODO: 2026-03-05 WE ARE HERE, WORKING ON THIS FUNCTION!! SciLifeLab rulez!


    def optimise(self):

        self.setup_rng()

        # synapse_density_override
        # n_synapess_override
        # synapse_position_override





    def write_log(self, text, flush=True):  # Change flush to False in future, debug
        if self.log_file is not None:
            self.log_file.write(text + "\n")

            if self.verbose:
                print(text)

            if flush:
                self.log_file.flush()
        else:
            if self.verbose:
                print(text)


    def load_trace_data(self, data_file):

        self.write_log(f"Loading {data_file}")

        with open(data_file, "r") as f:
            self.data = json.load(f)

            self.volt = np.array(self.data["data"]["mean_norm_trace"]).flatten()
            self.sample_freq = self.data["meta_data"]["sample_frequency"]

            if "holding_voltage" in self.data["meta_data"]:
                self.trace_holding_voltage = self.data["meta_data"]["trace_holding_voltage"]
            else:
                self.trace_holding_voltage = None

            dt = 1 / self.sample_freq
            self.time = 0 + dt * np.arange(0, len(self.volt))

            self.stim_time = np.array(self.data["meta_data"]["stim_time"])

            self.cell_type = self.data["meta_data"]["cell_type"]

    def save_parameter_data(self):

        if self.pc.id() != 0:
            self.write_log("No servants are allowed to write output to json, ignoring call.")
            return

        self.write_log(f"Saving data to {self.parameter_data_file_name}")
        self.synapse_parameter_data.save(self.parameter_data_file_name)

    def load_parameter_data(self):

        # TODO: How should this be done with new architecture?

        self.synapse_parameter_data = ParameterBookkeeper(old_book_file=self.parameter_data_file_name, n_max=100)
        self.synapse_parameter_data.check_integrity()

        best_dataset = self.synapse_parameter_data.get_best_dataset()

        if best_dataset is not None:
            self.synapse_section_id = best_dataset["section_id"]
            self.synapse_section_x = best_dataset["section_x"]

        if self.pc.id() != 0:
            # This is to prevent duplicating entries
            self.synapse_parameter_data.clear()



    def get_cell_properties(self):

        if self.cell_properties is None:
            with open(self.neuron_set_file, 'r') as f:
                self.cell_properties = json.load(f)

        cell_type = self.data["meta_data"]["cell_type"]

        return copy.deepcopy(self.cell_properties[cell_type])


    def update_cell_properties(self, holding_current):

        cell_type = self.data["meta_data"]["cell_type"]

        with open(self.neuron_set_file, 'r') as f:
            self.cell_properties = json.load(f)

        self.cell_properties[cell_type]["holding_current"] = holding_current

        with open(self.neuron_set_file, 'w') as f:
            json.dump(self.cell_properties, f, indent=4)


    def setup_model(self,
                    synapse_density_override=None,
                    n_synapses_override=None,
                    synapse_position_override=None):

        self.write_log(f"setup_model: synapse_position-override: {synapse_position_override}")

        params = {}

        t_stim = self.stim_time

        # Read the info needed to setup the neuron hosting the synapses
        c_prop = self.get_cell_properties()

        if synapse_position_override is not None:
            synapse_section_id, synapse_section_x = synapse_position_override
        else:
            synapse_section_id, synapse_section_x = None, None

        if synapse_density_override is not None:
            synapse_density = synapse_density_override
        else:
            synapse_density = c_prop["synapse_density"]

        if n_synapses_override is not None:
            n_synapses = n_synapses_override
        else:
            if "num_synapses" in c_prop:
                n_synapses = c_prop["num_synapses"]
            elif 'nSynapses' in c_prop:
                n_synapses = c_prop['nSynapses']
            else:
                raise Exception('Setup error: number of synapses not no specified in setup file (which ever that is?)')

        if "holding_current" in c_prop:
            holding_current = c_prop["holding_current"]
        else:
            holding_current = None

        neuron_morphology_key = c_prop["neuron_morphology_key"]
        neuron_parameter_key = c_prop["neuron_parameter_key"]

        # Use the trace holding voltage if it exists, otherwise use the holding voltage in the neuronSet json file.
        if self.trace_holding_voltage is not None:
            trace_holding_voltage = self.trace_holding_voltage
        elif "baseline_voltage" in c_prop:
            trace_holding_voltage = c_prop["baseline_voltage"]
            self.trace_holding_voltage = trace_holding_voltage
        else:
            trace_holding_voltage = None
            assert f"You need to specify either a trace_holding_voltage in {self.data_file}" \
                   f"or specify baselineVoltage in neuronSet.json for the neuron type in question."

        # Temporarily force regeneration of holding current
        holding_current = None

        print(f"Using random seed {self.seed}")

        self.rsr_synapse_model = \
            RunSynapseRun(neuron_path=snudda_parse_path(c_prop["neuron_path"], self.snudda_data),
                          neuron_morphology_key=neuron_morphology_key,
                          neuron_parameter_key=neuron_parameter_key,
                          stim_times=t_stim,
                          num_synapses=n_synapses,
                          synapse_density=synapse_density,
                          holding_voltage=trace_holding_voltage,
                          holding_current=holding_current,
                          synapse_type=self.synapse_type,
                          params=params,
                          time=self.sim_time,
                          log_file=self.log_file,
                          synapse_section_id=synapse_section_id,
                          synapse_section_x=synapse_section_x,
                          random_seed=self.seed,
                          verbose=True)

        if self.rsr_synapse_model.holding_current != holding_current:
            self.update_cell_properties(holding_current=self.rsr_synapse_model.holding_current)

        return self.rsr_synapse_model


    def get_peak_idx(self, stim_time, time, volt):

        freq = 1.0 / (stim_time[1] - stim_time[0])

        p_window = 1.0 / (2 * freq) * np.ones(stim_time.shape)
        p_window[-1] *= 5

        peak_info = self.find_peaks_helper(p_time=stim_time,
                                           p_window=p_window,
                                           time=time,
                                           volt=volt)

        return peak_info["peakIdx"]


    # Find peaks within pStart[i] and pStart[i]+pWindow[i]
    # The value is not the amplitude of the peak, just the voltage at the peak

    def find_peaks_helper(self, p_time, p_window, time=None, volt=None):

        peak_idx = []
        peak_time = []
        peak_volt = []

        for pt, pw in zip(p_time, p_window):
            t_start = pt
            t_end = pt + pw

            t_idx = np.where(np.logical_and(t_start <= time, time <= t_end))[0]

            if len(t_idx) == 0:
                self.write_log(f"No time points within {t_start} and {t_end}", flush=True)
                import pdb
                pdb.set_trace()

            assert len(t_idx) > 0, f"No time points within {t_start} and {t_end}"

            if self.synapse_type == "glut":
                p_idx = t_idx[np.argmax(volt[t_idx])]
            elif self.synapse_type == "gaba":
                # We assume that neuron is more depolarised than -65, ie gaba is
                # also depolarising
                p_idx = t_idx[np.argmax(volt[t_idx])]
            else:
                self.write_log("Unknown synapse type : " + str(self.synapse_type), flush=True)
                import pdb
                pdb.set_trace()

            peak_idx.append(int(p_idx))
            peak_time.append(time[p_idx])
            peak_volt.append(volt[p_idx])

        # Save to cache -- obs peakVolt is NOT amplitude of peak, just volt

        peak_dict = {"peakIdx": np.array(peak_idx),
                    "peakTime": np.array(peak_time),
                    "peakVolt": np.array(peak_volt)}  # NOT AMPLITUDE

        return peak_dict

    def find_trace_heights(self, time, volt, peak_idx):

        decay_func = lambda x, a, b, c: a * np.exp(-x / b) + c

        v_base = np.mean(volt[int(0.3 * peak_idx[0]):int(0.8 * peak_idx[0])])

        peak_height = np.zeros((len(peak_idx)))
        peak_height[0] = volt[peak_idx[0]] - v_base

        decay_fits = []

        for idx_b in range(1, len(peak_idx)):

            if peak_height[0] > 0:
                if idx_b < len(peak_idx) - 1:
                    p0d = [0.06, 0.05, self.trace_holding_voltage]
                else:
                    p0d = [1e-5, 100, self.trace_holding_voltage]

                    if self.synapse_type == "gaba":
                        p0d = [1e-8, 10000, self.trace_holding_voltage]
            else:
                # In some cases for GABA we had really fast decay back
                if idx_b < len(peak_idx) - 1:
                    p0d = [-0.06, 0.05, self.trace_holding_voltage]
                else:
                    p0d = [-1e-5, 1e5, self.trace_holding_voltage]

            peak_idx_a = peak_idx[idx_b - 1]  # Prior peak
            peak_idx_b = peak_idx[idx_b]  # Next peak

            if idx_b < len(peak_idx) - 1:
                # Not the last spike
                idx_start = int(peak_idx_a * 0.9 + peak_idx_b * 0.1)
                idx_end = int(peak_idx_a * 0.1 + peak_idx_b * 0.9)
            else:
                # Last spike, use only last half of decay trace
                idx_start = int(peak_idx_a * 0.5 + peak_idx_b * 0.5)
                idx_end = int(peak_idx_a * 0.1 + peak_idx_b * 0.9)  # might need 0.85 as last

            try:
                assert idx_start < idx_end
            except:
                import traceback
                tstr = traceback.format_exc()
                self.write_log(tstr, flush=True)

                import matplotlib.pyplot as plt

                plt.figure()
                plt.plot(time, volt)
                plt.xlabel("Time (error plot)")
                plt.ylabel("Volt (error plot)")
                plt.ion()
                plt.show()
                plt.title("ERROR!!!")
                import pdb
                pdb.set_trace()

            t_ab = time[idx_start:idx_end]
            v_ab = volt[idx_start:idx_end]

            t_ab_fit = t_ab - t_ab[0]
            v_ab_fit = v_ab

            try:

                try:
                    fit_params, pcov = scipy.optimize.curve_fit(decay_func, t_ab_fit, v_ab_fit, p0=p0d)
                except:
                    import traceback
                    tstr = traceback.format_exc()
                    self.write_log(tstr)

                    ### DEBUGGING START
                    import pickle
                    from datetime import datetime

                    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                    filename = f"curve_fit_args_{timestamp}.pkl"

                    dump_data = {
                        "t_ab_fit": t_ab_fit,
                        "v_ab_fit": v_ab_fit,
                        "p0d": p0d,
                    }

                    with open(filename, "wb") as f:
                        pickle.dump(dump_data, f)

                    print(f"Saved to {filename}")

                    ### DEBIGGING END

                    self.write_log("!!! Failed to converge, trying with smaller decay constant")
                    p0d[1] *= 0.01
                    fit_params, pcov = scipy.optimize.curve_fit(decay_func, t_ab_fit, v_ab_fit, p0=p0d)

                t_b = time[peak_idx_b] - t_ab[0]
                v_base_b = decay_func(t_b, fit_params[0], fit_params[1], fit_params[2])

                peak_height[idx_b] = volt[peak_idx_b] - v_base_b

                v_fit = decay_func(t_ab - t_ab[0], fit_params[0], fit_params[1], fit_params[2])
                decay_fits.append((t_ab, v_fit))

            except:
                self.write_log("Check that the threshold in the peak detection before is OK")
                # self.plot(name)
                import traceback
                tstr = traceback.format_exc()
                self.write_log(tstr, flush=True)

                if True:
                    import matplotlib.pyplot as plt

                    plt.figure()
                    plt.plot(t_ab, v_ab, 'r')
                    plt.title("Error in findTraceHeights")
                    plt.xlabel("time")
                    plt.ylabel("volt")
                    # plt.plot(tAB,vFit,'k-')
                    plt.ion()
                    plt.show()

                import pdb
                pdb.set_trace()

        return peak_height.copy(), decay_fits, v_base

    def neuron_synapse_helper_glut(self, t_spike,
                                   u, tau_r, tau_f, tau_ratio, cond,
                                   smooth_exp_trace8, smooth_exp_trace9, exp_peak_height,
                                   return_type="peaks"):

        if self.debug_pars_flag:
            self.debug_pars.append([u, tau_r, tau_f, tau_ratio, cond])

        params = self.synapse_parameters
        tau = tau_r * tau_ratio

        peak_h, t_sim, v_sim = self.run_model(t_spike, u,
                                              tau_r, tau_f, cond, tau,
                                              params=params,
                                              return_trace=True)

        # Calculate error in decay fit
        sim_trace8, sim_time8 = self.smoothing_trace(v_sim, self.num_smoothing,
                                                     time=t_sim,
                                                     start_time=self.decay_start_fit8,
                                                     end_time=self.decay_end_fit8)

        sim_trace9, sim_time9 = self.smoothing_trace(v_sim, self.num_smoothing,
                                                     time=t_sim,
                                                     start_time=self.decay_start_fit9,
                                                     end_time=self.decay_end_fit9)

        # We only want to use the bit of the trace after max
        idx_max8 = np.argmax(smooth_exp_trace8)
        idx_max9 = np.argmax(smooth_exp_trace9)

        # Calculating error in peak height
        if self.normalise_trace:
            h_diff = np.abs(peak_h/peak_h[0] - exp_peak_height/exp_peak_height[0])
        else:
            h_diff = np.abs(peak_h - exp_peak_height)

        h_diff[0] *= 3
        h_diff[-2] *= 2
        h_diff[-1] *= 3

        # This is to prevent the model spiking
        spike_penalty = np.sum(peak_h > 0.03) * 20

        h_error = np.sum(h_diff) / len(h_diff)

        decay_error8 = np.mean((smooth_exp_trace8[idx_max8:] - sim_trace8[idx_max8:]) ** 2)
        decay_error9 = np.mean((smooth_exp_trace9[idx_max9:] - sim_trace9[idx_max9:]) ** 2)

        fit_error = h_error + decay_error8 + decay_error9 + spike_penalty

        print(f"{fit_error = }")

        if spike_penalty > 0:
            self.write_log("Action potential detected in trace. Penalising!")

        if False:
            peak_base = v_sim[-1]
            plt.figure()
            plt.plot(t_sim, v_sim, 'k-')
            plt.plot(sim_time8, sim_trace8, 'y--')
            plt.plot(sim_time8, smooth_exp_trace8, 'r--')
            plt.plot(sim_time9, sim_trace9, 'y--')
            plt.plot(sim_time9, smooth_exp_trace9, 'r--')

            for tp, expH, modH in zip(t_spike, exp_peak_height, peak_h):
                plt.plot([tp, tp], [peak_base, expH + peak_base], 'r-', linewidth=3)
                plt.plot([tp, tp], [peak_base, modH + peak_base], 'b-')
            plt.title("hE = %g, dE8 = %g, dE9 = %g" \
                      % (h_error, decay_error8, decay_error9))

            plt.ion()
            plt.show()

        if return_type == "peaks":
            return peak_h
        elif return_type == "error":
            return fit_error
        elif return_type == "full":
            return fit_error, peak_h, t_sim, v_sim
        else:
            assert False, "Unknown return type: " + str(return_type)

    def run_model(self, t_spike, u, tau_r, tau_f, cond, tau,
                  params=None,
                  return_trace=False):

        if params is None:
            params = {}

        # self.writeLog("Running neuron model")

        assert self.rsr_synapse_model is not None, \
            "!!! Need to call setupModel first"

        # Should we make a copy of params, to not destroy it? ;)
        params["U"] = u
        params["tauR"] = tau_r
        params["tauF"] = tau_f
        params["cond"] = cond
        params["tau"] = tau

        # self.writeLog("params=" + str(params))

        self.pc.barrier()
        (t_sim, v_sim, i_sim) = self.rsr_synapse_model.run2(pars=params)

        if t_sim.shape != v_sim.shape:
            self.write_log("Shape are different, why?!", flush=True)
            import pdb
            pdb.set_trace()

        peak_idx = self.get_peak_idx(time=t_sim, volt=v_sim, stim_time=t_spike)
        peak_height, decay_fits, v_base = self.find_trace_heights(t_sim, v_sim, peak_idx)

        if return_trace:
            return peak_height, t_sim, v_sim
        else:
            return peak_height

    # This should read from a JSON file instead

    def get_model_bounds(self):

        mb = self.data["model_data"]

        param_list = ["U", "tauR", "tauF", "tauRatio", "cond"]
        lower_bound = [mb[x][0] for x in param_list]
        upper_bound = [mb[x][1] for x in param_list]

        return lower_bound, upper_bound


    # TODO: BUtcher this function
    def parallel_optimise_single_cell(self, n_trials=10000, post_opt=False):

        start_time = timeit.default_timer()

        if self.pc.id() != 0:
            self.write_log("parallel_optimise_single_cell should only be called on master node")
            return

        # 1. Setup workers
        params = self.synapse_parameters

        if self.synapse_section_id is not None:
            syn_override = self.synapse_section_id, self.synapse_section_x
        else:
            syn_override = None

        # 2. Setup one cell to optimise, randomise synapse positions
        synapse_model = self.setup_model(params=params,
                                         synapse_position_override=syn_override)

        # (volt,time) = self.getData(dataType,cellID)
        peak_idx = self.get_peak_idx(stim_time=self.stim_time,
                                     time=self.time,
                                     volt=self.volt)
        t_spikes = self.time[peak_idx]

        sigma = np.ones(len(peak_idx))
        sigma[-1] = 1. / 3

        peak_height, decay_fits, v_base = self.find_trace_heights(self.time, self.volt, peak_idx)

        if False:
            plt.figure()
            plt.plot(self.time, self.volt, color="black")
            for idx in range(len(decay_fits)):
                plt.plot(decay_fits[idx][0], decay_fits[idx][1], color="red")
            plt.show()

        skip_sets = self.synapse_parameter_data.old_iter
        print(f"Starting sobol sequence at position {skip_sets}")

        # 2b. Create list of all parameter points to investigate
        model_bounds = self.get_model_bounds()
        parameter_points = self.setup_parameter_set(model_bounds, n_trials, skip_sets=skip_sets)

        # 3. Send synapse positions to all workers, and split parameter points
        #    between workers

        if self.d_view is not None:
            self.setup_parallel(self.d_view)

            self.d_view.scatter("parameter_points", parameter_points, block=True)

            self.d_view.push({"params": params,
                              "synapse_section_id": synapse_model.synapse_section_id,
                              "synapse_section_x": synapse_model.synapse_section_x,
                              "model_bounds": model_bounds,
                              "stim_time": self.stim_time,
                              "peak_height": peak_height},
                             block=True)

            cmd_str_setup = \
                "ly.sobol_worker_setup(params=params," \
                + "synapse_position_override=(synapse_section_id,synapse_section_x))"

            self.d_view.execute(f"ly.write_log('TESTING LOG A')", block=True)

            self.write_log("Calling sobol_worker_setup")
            self.d_view.execute(cmd_str_setup, block=True)

            self.d_view.execute(f"ly.write_log('TESTING LOG B')", block=True)

            cmd_str = ("res = ly.sobol_scan(t_stim = stim_time,"
                       "                    h_peak = peak_height,"
                       "                    parameter_sets=parameter_points,"
                       "                    model_bounds=model_bounds,"
                       "                    smooth_exp_trace8=ly.smooth_exp_volt8,"
                       "                    smooth_exp_trace9=ly.smooth_exp_volt9)")

            self.write_log("Executing workers, bang bang")
            self.d_view.execute(cmd_str, block=True)

            # 5. Gather worker data
            self.write_log("Gathering results from workers")
            # res = self.d_view["res"]
            res = self.d_view.gather("res", block=True)
            self.write_log("Results gathered.")

            #  for r in res:
            self.synapse_parameter_data.merge(res)

            self.save_parameter_data()

        else:

            # No dView, run in serial mode...
            self.sobol_worker_setup(params=params,
                                    synapse_position_override=(synapse_model.synapse_section_id,
                                                               synapse_model.synapse_section_x))

            self.sobol_scan(parameter_sets=parameter_points,
                            t_stim=self.stim_time,
                            h_peak=peak_height,
                            model_bounds=model_bounds,
                            smooth_exp_trace8=self.smooth_exp_volt8,
                            smooth_exp_trace9=self.smooth_exp_volt9)

            self.save_parameter_data()


        self.write_log(f"Sobol search done. Best parameter {self.synapse_parameter_data.get_best_parameterset()}")

        if post_opt:
            # This updates parameters and saves new parameter cache
            self.get_refined_parameters()

            self.save_parameter_data()

        end_time = timeit.default_timer()

        self.write_log(f"Optimisation duration: {end_time - start_time}.1f s", flush=True)



if __name__ == "__main__":
    so = SynapseOptimiser()


"""

from skopt import gp_minimize
from skopt import Optimizer
from joblib import Parallel, delayed


## Parallell optimisation

opt = Optimizer(dimensions=m_bounds, random_state=42)

n_iterations = 10
batch_size = 8  # tune this to your number of CPU cores

# TODO: This can not handle the self reference, need to make func
#       self contained.

for i in range(n_iterations):
    print(f"Iteration {i}/{n_iterations}")
    x_batch = opt.ask(n_points=batch_size)

    # Here run simulation, get results and put it in y_batch

    # y_batch = Parallel(n_jobs=-1)(delayed(func)(x) for x in x_batch)  # couldnt pickle
    opt.tell(x_batch, y_batch)

best_idx = opt.yi.index(min(opt.yi))
print("Best value:", min(opt.yi))
print("Best params:", opt.Xi[best_idx])

fit_params = opt.Xi[best_idx]
min_error = opt.yi
"""


import unittest
import os
import sys
import json
import numpy as np

from snudda import SnuddaLoad

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from snudda.init.init import SnuddaInit
from snudda.place.place import SnuddaPlace


class TestPlace(unittest.TestCase):

    def setUp(self):
        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        print(f"Current directory (detect): {os.path.dirname(os.path.realpath(__file__))}")

        neuron_dir = os.path.join(os.path.dirname(__file__), "validation")
        self.sim_name = os.path.join("networks", "network_testing_place")
        self.config_file = os.path.join(self.sim_name, "network-config.json")
        cnc = SnuddaInit(struct_def={}, config_file=self.config_file, random_seed=1234)
        cnc.define_striatum(num_dSPN=10, num_iSPN=0, num_FS=10, num_LTS=0, num_ChIN=0,
                            volume_type="cube", neurons_dir=neuron_dir)
        cnc.write_json(self.config_file)

    def tearDown(self):
        pass

    def test_init(self):
        with open(self.config_file, "r") as f:
            config_data = json.load(f)

        self.assertTrue("random_seed" in config_data)
        self.assertTrue(config_data["random_seed"]["master_seed"] == 1234)
        self.assertTrue(config_data["random_seed"]["init"] == 2906597030)
        self.assertTrue(config_data["random_seed"]["place"] == 1602421836)
        self.assertTrue(config_data["random_seed"]["detect"] == 216676975)
        self.assertTrue(config_data["random_seed"]["project"] == 2698621808)
        self.assertTrue(config_data["random_seed"]["prune"] == 507409703)
        self.assertTrue(config_data["random_seed"]["input"] == 2825158027)
        self.assertTrue(config_data["random_seed"]["simulate"] == 3613074)

        self.assertTrue("regions" in config_data)
        self.assertTrue("Striatum" in config_data["regions"])

        self.assertTrue("connectivity" in config_data["regions"]["Striatum"])
        self.assertTrue("neurons" in config_data["regions"]["Striatum"])

        for neuron in config_data["regions"]["Striatum"]["neurons"]:
            self.assertTrue("neuron_path" in config_data["regions"]["Striatum"]["neurons"][neuron])
            self.assertTrue("num_neurons" in config_data["regions"]["Striatum"]["neurons"][neuron])
            self.assertTrue("neuron_type" in config_data["regions"]["Striatum"]["neurons"][neuron])
            self.assertTrue("volume_id" in config_data["regions"]["Striatum"]["neurons"][neuron])

    def test_place(self):

        # Place neurons

        position_file = os.path.join(self.sim_name, "network-neuron-positions.hdf5")

        npn = SnuddaPlace(config_file=self.config_file,
                          log_file=None,
                          verbose=True,
                          d_view=None,          # TODO: If d_view is None code run sin serial, add test parallel
                          h5libver="latest")

        npn.parse_config()
        npn.write_data(position_file)

        with self.subTest(stage="neuron_count"):
            num_cells = 20
            self.assertEqual(npn.all_neuron_positions().shape[0], num_cells)
            self.assertEqual(len(npn.neurons), num_cells)

        with self.subTest(stage="d_min"):
            # Check that minimum distance between neurons, d_min is fulfilled
            all_pos = npn.all_neuron_positions()
            for pos in all_pos:
                d_min = 1e-5
                # Not too close (self comparison allowed to be equal, hence -1)
                self.assertTrue(np.sum(np.linalg.norm(all_pos - pos, axis=1) >= d_min) == all_pos.shape[0]-1)

                # Not too far apart
                d_max = 1e4  # This is just for this simulation, to catch if there are some spurious neurons
                self.assertTrue(np.sum(np.linalg.norm(all_pos - pos, axis=1) <= d_max) == all_pos.shape[0])

        # TODO: Load hdf5 file and check that data is what we expect

    def test_population_units(self, stage="place-pop-unit-random"):

        network_path = os.path.join(os.path.dirname(__file__), "networks", "network_place_pop_unit_random")
        cnc = SnuddaInit(struct_def={}, network_path=network_path, random_seed=123457)
        cnc.define_striatum(num_dSPN=1000, num_iSPN=1000, num_FS=20, num_LTS=0, num_ChIN=0,
                            volume_type="cube")
        cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN", "iSPN"],
                                       fraction_of_neurons=0.5)
        cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN", "iSPN"],
                                       fraction_of_neurons=0.2)
        cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN"],
                                       fraction_of_neurons=0.3)
        cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["iSPN"],
                                       fraction_of_neurons=0.15)
        cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["iSPN"],
                                       fraction_of_neurons=0.15, unit_id=10)
        cnc.write_json()

        npn = SnuddaPlace(network_path=network_path, h5libver="latest", verbose=True)
        npn.parse_config()
        npn.write_data()

        network_file = os.path.join(network_path, "network-neuron-positions.hdf5")
        sl = SnuddaLoad(network_file)

        pop_units = sl.get_neuron_population_units()
        neuron_types = sl.get_neuron_types()

        self.assertTrue(np.abs(np.sum(pop_units == 1) - 2000*0.5) < 50)  # 50% of dSPN, iSPN should be pop unit 1
        self.assertTrue(np.abs(np.sum(pop_units == 2) - 2000*0.2) < 50)  # 20% should be pop unit 1
        self.assertTrue(np.abs(np.sum(pop_units == 3) - 1000*0.3) < 35)  # 30% of dSPN should be pop unit 1
        self.assertTrue(np.abs(np.sum(pop_units == 4) - 1000*0.15) < 35)  # 15% of iSPN should be pop unit 1
        self.assertTrue(np.abs(np.sum(pop_units == 10) - 1000*0.15) < 35)  # 15% of iSPN should be pop unit 1


if __name__ == '__main__':
    unittest.main()

# python3 -m unittest test_place



import os
import unittest
import numpy as np
from snudda.place.rotation import SnuddaRotate


class TestRotation(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        self.sr = SnuddaRotate()
        self.sr.parse_config_file("data/RotationTest.json")

    def test_rotations(self):

        rng = np.random.default_rng()

        rot = self.sr.get_rotations(volume_name="Striatum", neuron_type="LTS",
                                    neuron_positions=[[0, 0, 0]],
                                    rng=rng)
        self.assertEqual(np.linalg.det(rot[0]), 1)

        vec_test = np.array([0, 0, 1])
        vec_rot = rot[0].dot(vec_test)
        vec_facit = np.array([-1, -1, -1])
        self.assertTrue(np.allclose(vec_rot / np.linalg.norm(vec_rot), vec_facit / np.linalg.norm(vec_facit)))

        rot2 = self.sr.get_rotations(volume_name="Striatum", neuron_type="LTS",
                                     neuron_positions=[[0, 0, 0.5e-3]],
                                     rng=rng)
        self.assertTrue(np.abs(np.linalg.det(rot2[0]) - 1) < 1e-6)

        vec_rot2 = rot2[0].dot(vec_test)
        vec_facit2 = np.array([-1, -1, 0])

        self.assertTrue(np.allclose(vec_rot2 / np.linalg.norm(vec_rot2),
                                    vec_facit2 / np.linalg.norm(vec_facit2),
                            atol=1e-6))

        # TODO: Also need to test the random rotations


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import numpy as np
import copy
from scipy.spatial.distance import pdist, squareform


class PointCluster:

    def __init__(self):
        self.coords = None
        self.dist = None
        self.num = 0

    def generate_cube(self, centre, side, num):
        self.coords = side * np.random.uniform(size=(num, 3))
        self.coords[:, 0] += centre[0]
        self.coords[:, 1] += centre[1]
        self.coords[:, 2] += centre[2]

        self.num = num
        self.calc_distance()

    def generate_disk(self, centre, side, num):
        self.coords = side * np.random.uniform(size=(num, 3))
        self.coords[:, 0] += centre[0]
        self.coords[:, 1] += centre[1]
        self.coords[:, 2] = centre[2]

        self.num = num
        self.calc_distance()

    def calc_distance(self):
        self.dist = squareform(pdist(self.coords))

    def get_colour(self, idx):
        min_coord = np.min(self.coords)
        max_coord = np.max(self.coords)

        return np.divide(self.coords[idx, :] - min_coord, max_coord - min_coord)


class PointMap:

    def __init__(self, source, dest):

        self.source = source
        self.dest = dest

        self.map = np.arange(source.num)

        assert source.num == dest.num, f"source and dest should have equal number of points"

        self.start_energy = None

    def delta_energy_WRONG(self, a, b):

        # For each pair of neurons in the source structure, find their distance, and the distance of their projections.
        # The sum of the product of these distances is the total energy.
        # So the question is how does this sum change if we would swap the neurons.

        a_dest = self.map[a]
        b_dest = self.map[b]

        test_map = copy.deepcopy(self.map)
        test_map[a] = self.map[b]
        test_map[b] = self.map[a]

        energy_before = np.sum(np.multiply(self.source.dist[a, :], self.dest.dist[a_dest, :][test_map])) \
                        + np.sum(np.multiply(self.source.dist[b, :], self.dest.dist[b_dest, :][test_map]))

        energy_after = np.sum(np.multiply(self.source.dist[a, :], self.dest.dist[b_dest, :][test_map])) \
                       + np.sum(np.multiply(self.source.dist[b, :], self.dest.dist[a_dest, :][test_map]))

        return energy_after - energy_before

    def delta_energy(self, a, b):

        energy_before = self.energy()
        alt_map = copy.deepcopy(self.map)
        alt_map[a] = self.map[b]
        alt_map[b] = self.map[a]
        energy_after = self.energy(alt_map)

        return energy_after - energy_before

    def energy(self, alt_map=None):

        if alt_map is not None:
            return np.sum(np.power(self.source.dist - self.dest.dist[:, alt_map][alt_map, :], 2))

        return np.sum(np.power(self.source.dist - self.dest.dist[:, self.map][self.map, :], 2))

    def iterate(self, n_iter):

        if self.start_energy is None:
            self.start_energy = self.energy()

        # Pick a pair of points in source, check if their dest points should be swapped
        ab = np.random.randint(low=0, high=self.source.coords.shape[0], size=(n_iter, 2))
        p_val = np.random.uniform(size=n_iter)

        for idx, ((a, b), px) in enumerate(zip(ab, p_val)):
            energy_change = self.delta_energy(a, b)
            p_swap = 1 / (1 + np.exp(4 * energy_change / self.start_energy * 100))
            print(f"{idx}/{n_iter}  energy = {self.energy()}, p_swap = {p_swap}")
            if px < p_swap:
                # Swap the two if energy
                old_a = self.map[a]
                self.map[a] = self.map[b]
                self.map[b] = old_a

            if idx % 100000 == 0:
                self.plot_map()

        print(f"Energy = {self.energy()}")
        self.plot_map()

    def plot_map(self):

        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        for idx in range(self.source.num):
            col = self.source.get_colour(idx)
            sp = self.source.coords[idx, :]
            dp = self.dest.coords[self.map[idx], :]
            ax.plot(xs=sp[0], ys=sp[1], zs=sp[2], color=col, marker='o')
            ax.plot(xs=dp[0], ys=dp[1], zs=dp[2], color=col, marker='o')

        fig.show()
        plt.pause(0.1)


if __name__ == "__main__":
    pc_a = PointCluster()
    pc_b = PointCluster()

    pc_a.generate_disk(centre=np.array([0, 0, 0]), side=1000e-6, num=100)
    pc_b.generate_disk(centre=np.array([0, 0, 5e-3]), side=1000e-6, num=100)

    pm = PointMap(source=pc_a, dest=pc_b)

    pm.iterate(10000)

    import pdb

    pdb.set_trace()



import os
from collections import OrderedDict

import numpy as np

from snudda.utils.load import SnuddaLoad

import matplotlib.pyplot as plt

from snudda.utils.load_network_simulation import SnuddaLoadSimulation


class SnuddaPlotSpikeRaster2:

    def __init__(self, network_path, network_file=None, simulation_file=None, figure_path=None,
                 snudda_load=None, snudda_simulation_load=None):

        """ If you pass snudda_load and snudda_simulation_load those will be used, so you do not
            have to reload the data if you do multiple types of plots. """

        if network_path is not None:
            self.network_path = network_path
        elif snudda_load is not None:
            self.network_path = os.path.dirname(snudda_load.network_file)
        else:
            self.network_path = None

        if network_file:
            self.network_file = network_file
        elif network_path is not None:
            self.network_file = os.path.join(self.network_path, "network-synapses.hdf5")
        elif snudda_load:
            self.network_file = snudda_load.network_file
        else:
            self.network_file = None

        if simulation_file:
            self.simulation_file = simulation_file
        else:
            self.simulation_file = os.path.join(self.network_path, "simulation", "output.hdf5")

        if figure_path:
            self.figure_path = figure_path
        else:
            self.figure_path = os.path.join(self.network_path, "figures")

        if snudda_load:
            self.snudda_load = snudda_load
            assert network_file is None or snudda_load.network_file == self.network_file, \
                f"snudda_load refers to {snudda_load.network_file}, but user passed network_file={self.network_file}"
        else:
            self.snudda_load = SnuddaLoad(network_file=self.network_file, load_synapses=False)

        if snudda_simulation_load:
            self.snudda_simulation_load = snudda_simulation_load
            assert simulation_file is None \
                   or snudda_simulation_load.network_simulation_output_file_name == simulation_file, \
                f"snudda_simulation_load refers to {snudda_simulation_load.self.network_simulation_output_file_name}," \
                f" but user passed simulation_file={simulation_file}"
        else:
            self.snudda_simulation_load = SnuddaLoadSimulation(network_simulation_output_file=self.simulation_file)

        spike_data = self.snudda_simulation_load.merge_spikes()

        self.spike_time = spike_data[:, 0]
        self.spike_neuron_id = spike_data[:, 1].astype(int)
        self.traces = self.snudda_simulation_load.get_voltage()
        self.time = self.snudda_simulation_load.get_time()

    def make_figures_directory(self):

        if not os.path.isdir(self.figure_path):
            os.makedirs(self.figure_path)

    @staticmethod
    def get_colours(neuron_type):

        # TODO: Read colours from a JSON file

        colours = {"dSPN".lower(): (77. / 255, 151. / 255, 1.0),
                   "iSPN".lower(): (67. / 255, 55. / 255, 181. / 255),
                   "FS".lower(): (6. / 255, 31. / 255, 85. / 255),
                   "FSN".lower(): (6. / 255, 31. / 255, 85. / 255),
                   "ChIN".lower(): (252. / 255, 102. / 255, 0.0),
                   "LTS".lower(): (150. / 255, 63. / 255, 212. / 255)}

        if neuron_type.lower() in colours:
            return colours[neuron_type.lower()]
        else:
            return (0, 0, 0)

    def get_all_colours(self):

        neuron_type_list = self.snudda_load.get_neuron_types(return_set=False)
        neuron_colours = np.zeros((len(neuron_type_list), 3))

        for idx, nt in enumerate(neuron_type_list):
            neuron_colours[idx, :] = self.get_colours(nt)

        return neuron_colours

    # TODO: Add background colour to population units
    def plot_hist_raster(self, type_order=None, skip_time=0, end_time=None, fig_size=None, type_division=None, fig_file=None):
        # type_division: divides plot in two parts based on neuron type.
        #   Example 1: [["dspn","ispn"],["chin", "fsn","lts"]]
        #   Example 2: [["dspn","ispn","chin", "fsn","lts"],[]]

        self.make_figures_directory()
        # Gets a list of all the neurons' types
        neuron_type_list = self.snudda_load.get_neuron_types(return_set=False)
        neuron_population_unit_list = self.snudda_load.get_neuron_population_units(return_set=False)

        neuron_type_map = dict()

        if end_time is None:
            end_time = 1.02 * max(self.time)
        if type_order is None:
            unique_neuron_types = sorted(list(set(neuron_type_list)))
            type_order = unique_neuron_types
        else:
            unique_neuron_types = type_order + sorted(list(set(neuron_type_list) - set(type_order)))

        if type_division is None:
            type_division = [unique_neuron_types, []]

        for nt_idx, nt in enumerate(unique_neuron_types):
            neuron_type_map[nt] = nt_idx

        # For each neuron, associate the number of the type it is
        neuron_type_idx = np.array([neuron_type_map[x] for x in neuron_type_list])

        neuron_order = np.lexsort((neuron_population_unit_list, neuron_type_idx))

        # neuron_order = np.argsort(neuron_type_idx)
        neuron_order_lookup = np.zeros(neuron_order.shape)

        for idx, no in enumerate(neuron_order):
            neuron_order_lookup[no] = idx

        # new dict with cell type specific spikes
        type_dict = {'order': type_order}
        for t in type_order:
            type_dict[t] = [i for i, x in enumerate(neuron_type_list) if x.lower() == t.lower()]

        plt.rcParams.update({'font.size': 24,
                             'xtick.labelsize': 20,
                             'ytick.labelsize': 20,
                             'legend.loc': 'best'})
            
        # Prepare figure
        if not fig_size:
            fig_size = (10, 10)
        fig = plt.figure(figsize=fig_size)
        plt.rcParams.update({'font.size': 22})

        r = 4
        grid = plt.GridSpec(r, r, hspace=0, wspace=0)
        ax = fig.add_subplot(grid[2:, :])
        if (len(type_division[0]) > 0) & (len(type_division[1]) > 0):
            atop = fig.add_subplot(grid[0, :])
            atop2 = fig.add_subplot(grid[1, :])
        elif (len(type_division[0]) == 0) & (len(type_division[1]) == 0):
            print("No neuron type to show in histogram due to empty variable type_division.")
        elif (len(type_division[0]) == 0) | (len(type_division[1]) == 0):
            atop = fig.add_subplot(grid[0:2, :])
            atop2 = atop

        # Plot raster plot
        spike_y = np.take(neuron_order_lookup, self.spike_neuron_id)
        colour_lookup = self.get_all_colours()
        sc = np.zeros((len(spike_y), 3))
        for i in range(0, 3):
            sc[:, i] = np.take(colour_lookup[:, i], self.spike_neuron_id)
        ax.scatter(self.spike_time - skip_time, spike_y, color=sc, s=5, linewidths=0.1)

        max0 = 0
        max1 = 0
        spike2type = np.take(neuron_type_list, self.spike_neuron_id)

        # Spike rates
        for t in type_order:
            cell_spike_rates = np.array([])
            spikes_of_type = [i for i, x in enumerate(spike2type) if x.lower() == t.lower()]
            pruned_spikes = self.spike_time[spikes_of_type] - skip_time
            pruned_spikes = pruned_spikes[pruned_spikes > 0]
            spike_y_t = spike_y[spikes_of_type]
            spike_y_t = spike_y_t[pruned_spikes > 0]
            uspike_y_t = np.unique(spike_y_t)

            for cellid in uspike_y_t:
                cell_spike_rate = len(pruned_spikes[cellid == spike_y_t]) / (end_time - skip_time)
                cell_spike_rates = np.append(cell_spike_rates, cell_spike_rate)
            num_of_type = len(type_dict[t])
            cell_spike_rates = np.append(cell_spike_rates, np.zeros(num_of_type - len(cell_spike_rates)))

            print("NeuronType: {0}, Mean: {1}Hz Stdev: {2}Hz".format(t, np.mean(cell_spike_rates),
                                                                     np.std(cell_spike_rates)))

        # histogram
        for t in type_order:
            spikes_of_type = [i for i, x in enumerate(spike2type) if x.lower() == t.lower()]
            pruned_spikes = self.spike_time[spikes_of_type] - skip_time
            pruned_spikes = pruned_spikes[pruned_spikes > 0]
            num_of_type = len(type_dict[t])  #
            nspikes = len(pruned_spikes)
            bin_width = 0.05  # 10  # ms
            bin_range = np.arange(0, end_time - skip_time + bin_width, bin_width)
            if (nspikes > 0) & (t.lower() in type_division[0]):
                counts0, bins0, bars0 = atop.hist(pruned_spikes,
                                                  bins=bin_range,
                                                  range=(skip_time, end_time),
                                                  density=0,
                                                  color=sc[spikes_of_type][0],
                                                  alpha=1.0,
                                                  histtype='step',
                                                  weights=np.ones_like(pruned_spikes) / bin_width / num_of_type)

                max0 = max(np.append(counts0, max0))
            elif (nspikes > 0) & (t.lower() in type_division[1]):
                counts1, bins1, bars1 = atop2.hist(pruned_spikes,
                                                   bins=bin_range,
                                                   range=(skip_time, end_time),
                                                   density=0,
                                                   color=sc[spike2type == t][0],
                                                   alpha=1.0,
                                                   histtype='step',
                                                   weights=np.ones_like(pruned_spikes) / bin_width / num_of_type)

                max1 = max(np.append(counts1, max1))
            spike_rate = len(pruned_spikes) / num_of_type / (end_time - skip_time)
            print("{0}: {1} Hz".format(t, spike_rate))
        # Get position of labels
        unique_neuron_types = set(neuron_type_list)
        y_tick = []
        y_tick_label = []
        for nt in unique_neuron_types:
            y_tick_label.append(nt)
            y_tick.append(np.mean(neuron_order_lookup[np.where([x == nt for x in neuron_type_list])[0]]))

        ax.invert_yaxis()
        ax.set_xlabel('Time (s)', fontsize=20)
        ax.set_yticks(y_tick, fontsize=20)
        ax.set_yticklabels(y_tick_label)

        if skip_time or end_time:
            x_lim = ax.get_xlim()
            x_lim = (0, x_lim[1])
            if end_time:
                x_lim = (x_lim[0], end_time)
            ax.set_xlim(x_lim)

        atop.set_xlim(x_lim)
        atop2.set_xlim(x_lim)
        atop.set_xticklabels([])
        atop.set_ylabel('Mean spikes/s')

        if fig_file is None:
            fig_file = os.path.join(self.figure_path, "spike-histogram-raster.pdf")
        else:
            fig_file = os.path.join(self.figure_path, fig_file)

        print(f"Writing figure to {fig_file}")
        plt.tight_layout()
        self.make_figures_directory()
        plt.savefig(fig_file, dpi=300)

        plt.ion()
        plt.show()

        return ax

    def calculate_period_synchrony(self, period, neuron_id=None, time_range=None):

        spikes = self.snudda_simulation_load.get_spikes(neuron_id=neuron_id)
        all_spikes = []

        for s in spikes.values():

            sf = s.flatten()

            if time_range is not None:
                idx = np.where(np.logical_and(time_range[0] <= sf, sf <= time_range[1]))[0]
                all_spikes = all_spikes + list(sf[idx] % period)
            else:
                all_spikes = all_spikes + list(sf % period)

        phases = 2*np.pi/period * np.array(all_spikes)
        x = np.sum(np.cos(phases))
        y = np.sum(np.sin(phases))
        vs = np.sqrt(x**2 + y**2) / phases.size

        # Verify this is correct

        return vs

    def plot_spike_histogram_type(self, neuron_type, time_range=None, bin_size=50e-3, fig_size=(4.5, 3.5),
                                  fig_file=None, label_text=None, show_figure=True, n_core=None,
                                  linestyle="-", line_colours=None, linewidth=3,
                                  legend_loc="best", bbox_anchor=None, ax=None):

        self.make_figures_directory()

        plt.rcParams.update({'font.size': 12,
                             'xtick.labelsize': 12,
                             'ytick.labelsize': 12})

        assert type(neuron_type) == list, "neuron_type should be a list of neuron types"

        all_spikes = OrderedDict()
        neurons_of_type = OrderedDict()

        if time_range is None:
            time_range = (0, self.snudda_simulation_load.get_time()[-1])

        for nt in neuron_type:
            if n_core:
                neuron_id = [x for x, y
                             in self.snudda_load.get_centre_neurons_iterator(neuron_type=nt, n_neurons=n_core)]
            else:
                neuron_id = self.snudda_load.get_neuron_id_of_type(nt)

            spikes = self.snudda_simulation_load.get_spikes(neuron_id=neuron_id)
            neurons_of_type[nt] = neuron_id
            all_spikes[nt] = self.snudda_simulation_load.merge_spikes(spikes)[:, 0]

        bins = np.arange(time_range[0], time_range[1], bin_size)
        weights = [np.full(y.shape, 1/(len(x)*bin_size)) for x, y in zip(neurons_of_type.values(), all_spikes.values())]

        if label_text is None:
            label_text = ""

        if ax is None:
            fig = plt.figure(figsize=fig_size)
            ax = fig.add_subplot()

        if len(all_spikes.keys()) > 1:
            all_labels = [f"{label_text}{x}" for x in all_spikes.keys()]
        elif label_text is not None:
            all_labels = [label_text]
        else:
            all_labels = None

        if line_colours is None:
            line_colours = [self.get_colours(x) for x in all_spikes.keys()]

        for (nt, spikes), w, col in zip(all_spikes.items(), weights, line_colours):
            counts, edges = np.histogram(
                spikes,
                bins=bins,
                weights=w
            )

            ax.step(
                edges[:-1],
                counts,
                where="post",
                linewidth=linewidth,
                linestyle=linestyle,
                color=col,
                label=f"{label_text}{nt}" if label_text else nt
            )

#        ax.hist(x=all_spikes.values(), bins=bins, weights=weights, linewidth=linewidth, linestyle=linestyle,
#                histtype="step", color=line_colours,
#                label=all_labels)

        ax.set_xlim(time_range)

        plt.xlabel("Time (s)", fontsize=12)
        plt.ylabel("Frequency (Hz)", fontsize=12)

        handles, labels = ax.get_legend_handles_labels()
        if handles:
            ax.legend(loc=legend_loc, bbox_to_anchor=bbox_anchor)


        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['bottom'].set_linewidth(0.8)

        plt.tight_layout()

        if fig_file:
            fig_path = os.path.join(self.figure_path, fig_file)
            print(f"Writing figure {fig_path}")
            self.make_figures_directory()

            plt.savefig(fig_path)

        if show_figure:
            plt.ion()
            plt.show()

        return ax

    def plot_spike_histogram(self, population_id=None, neuron_type=None,
                             skip_time=0, end_time=None, fig_size=(4.5, 3.5), bin_size=50e-3,
                             fig_file=None, ax=None, label_text=None, show_figure=True, save_figure=True, colour=None,
                             linestyle="-", legend_loc="best", title=None, bbox_anchor=None):

        if population_id is None:
            population_id = self.snudda_load.get_neuron_population_units(return_set=True)

        if neuron_type is not None:

            if type(neuron_type) != list:
                neuron_type = [neuron_type]

            keep_neuron_id = set()
            for nt in neuron_type:
                keep_neuron_id |= set(self.snudda_load.get_neuron_id_of_type(neuron_type=nt))

            print(f"Processing {len(keep_neuron_id)} neurons of type {neuron_type}")
        else:
            keep_neuron_id = None

        self.make_figures_directory()

        plt.rcParams.update({'font.size': 12,
                             'xtick.labelsize': 12,
                             'ytick.labelsize': 12},
                            )

        if ax is None:
            fig = plt.figure(figsize=fig_size)
            ax = fig.add_subplot()
        
        pop_members = dict()
        pop_spikes = dict()

        if np.issubdtype(type(population_id), np.integer):
            population_id = np.array([population_id])

        for pid in population_id:
            members = self.snudda_load.get_population_unit_members(pid)

            if keep_neuron_id is not None:
                members = np.array(list(set(members) & keep_neuron_id))

            if len(members) == 0:
                # No members, skip it
                continue

            pop_members[pid] = members

            spikes = self.snudda_simulation_load.get_spikes(pop_members[pid])
            pop_spikes[pid] = self.snudda_simulation_load.merge_spikes(spikes)[:, 0]

        if end_time is None:
            end_time = self.snudda_simulation_load.get_time()[-1]

        try:
            bins = np.arange(skip_time, end_time+bin_size/2, bin_size)
            weights = [np.full(y.shape, 1/(len(x)*bin_size)) for x, y in zip(pop_members.values(), pop_spikes.values())]
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        if label_text is None:
            label_text = ""

        if len(pop_spikes.keys()) > 1:
            label = [f"{label_text}{x}" for x in pop_spikes.keys()]
        else:
            label = None

        N, bins, patches = ax.hist(x=pop_spikes.values(), bins=bins, weights=weights, linewidth=3, linestyle=linestyle,
                                   histtype="step", color=colour,
                                   label=label)

        if type(colour) == list:
            for patch, col in zip(patches, colour):
                patch[0].set_facecolor(col)

        plt.xlabel("Time (s)", fontsize=12)
        plt.ylabel("Frequency (Hz)", fontsize=12)
        ax.legend(loc=legend_loc, bbox_to_anchor=bbox_anchor)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['bottom'].set_linewidth(0.8)

        if title:
            plt.title(title)

        if fig_file is None:
            fig_file = os.path.join(self.figure_path,
                                    f"spike-frequency-pop-units{'-'.join([f'{x}' for x in pop_members.keys()])}.png")
        else:
            fig_file = os.path.join(self.figure_path, fig_file)

        plt.tight_layout()

        if save_figure:
            print(f"Saving figure {fig_file}")
            self.make_figures_directory()

            plt.savefig(fig_file, dpi=300)

        if show_figure:
            plt.ion()
            plt.show()

        return ax

    # Use this to plot a histogram for an arbitrary group of neurons specified with neuron_id
    def plot_group_spike_histogram(self, neuron_id=None,
                                    skip_time=0, end_time=None, fig_size=None, bin_size=50e-3,
                                    fig_file=None, ax=None, label_text=None, show_figure=True, save_figure=True, colour=None,
                                    linestyle="-", legend_loc="best", title=None):

        self.make_figures_directory()

        plt.rcParams.update({'font.size': 12,
                             'xtick.labelsize': 12,
                             'ytick.labelsize': 12,
                             'legend.loc': legend_loc})

        if ax is None:
            fig = plt.figure(figsize=fig_size)
            ax = fig.add_subplot()

        spikes = self.snudda_simulation_load.get_spikes(neuron_id)
        spikes = self.snudda_simulation_load.merge_spikes(spikes)[:, 0]

        if end_time is None:
            end_time = self.snudda_simulation_load.get_time()[-1]

        bins = np.arange(skip_time, end_time + bin_size / 2, bin_size)
        weights = 1 / (len(neuron_id) * bin_size)

        try:
            N, bins, patches = ax.hist(x=spikes, bins=bins, weights=np.full(spikes.shape, weights), linewidth=3, linestyle=linestyle,
                                       histtype="step", color=colour,
                                       label=label_text)
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        if type(colour) == list:
            for patch, col in zip(patches, colour):
                patch[0].set_facecolor(col)

        plt.xlabel("Time (s)", fontsize=12)
        plt.ylabel("Frequency (Hz)", fontsize=12)

        if label_text is not None:
            ax.legend()

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['bottom'].set_linewidth(0.8)

        if title:
            plt.title(title)

        if fig_file is None:
            fig_file = os.path.join(self.figure_path,
                                    f"spike-frequency-for-group.pdf")
        else:
            fig_file = os.path.join(self.figure_path, fig_file)

        plt.tight_layout()

        if save_figure:
            self.make_figures_directory()

            print(f"Saving figure {fig_file}")
            plt.savefig(fig_file, dpi=300)

        if show_figure:
            plt.ion()
            plt.show()

        return ax

    def plot_spike_raster(self, type_order=None, skip_time=0, end_time=None, fig_size=(4.5, 3.5), fig_file=None,
                          time_range=None, title=None, sort_direction=None, marker_size=1.0):

        # You can use sort_direction = "x", "y" or "z" to sort within the neuron type if you want to show the in x-direction

        self.make_figures_directory()

        plt.rcParams.update({'font.size': 12,
                             'xtick.labelsize': 12,
                             'ytick.labelsize': 12,
                             'legend.loc': 'best'})
        
        fig = plt.figure(figsize=fig_size)
        ax = fig.add_subplot()

        # Gets a list of all the neurons' types
        neuron_type_list = self.snudda_load.get_neuron_types(return_set=False)
        neuron_population_unit_list = self.snudda_load.get_neuron_population_units(return_set=False)

        neuron_type_map = dict()

        if type_order is None:
            unique_neuron_types = sorted(list(set(neuron_type_list)))
        else:
            unique_neuron_types = type_order + sorted(list(set(neuron_type_list) - set(type_order)))

        for nt_idx, nt in enumerate(unique_neuron_types):
            neuron_type_map[nt] = nt_idx

        # For each neuron, associate the number of the type it is
        neuron_type_idx = np.array([neuron_type_map[x] for x in neuron_type_list])
        # neuron_order = np.argsort(neuron_type_idx)
        if sort_direction is None:
            neuron_order = np.lexsort((neuron_population_unit_list, neuron_type_idx))
        else:
            neuron_positions = np.vstack([x["position"] for x in self.snudda_load.data["neurons"]])

            pos_order = neuron_positions[:, {"x": 0, "y": 1, "z": 2}[sort_direction]]
            neuron_order = np.lexsort((neuron_population_unit_list, pos_order, neuron_type_idx))

        neuron_order_lookup = np.zeros(neuron_order.shape)

        # for idx, no in enumerate(neuron_order):
        #     neuron_order_lookup[no] = idx

        idx = 0
        for no in neuron_order:
            # Skip the virtual neurons
            if not self.snudda_load.data["neurons"][no]["virtual_neuron"]:
                neuron_order_lookup[no] = idx
                idx += 1

        spike_y = np.take(neuron_order_lookup, self.spike_neuron_id)

        colour_lookup = self.get_all_colours()
        sc = np.zeros((len(spike_y), 3))

        for i in range(0, 3):
            sc[:, i] = np.take(colour_lookup[:, i], self.spike_neuron_id)

        ax.scatter(self.spike_time - skip_time, spike_y, color=sc, s=marker_size, linewidths=0.1,
                   rasterized=True, alpha=0.8)

        # Optionally we should also show the virtual neuron spikes...


        # Get position of labels
        unique_neuron_types = set(neuron_type_list)
        y_tick = []
        y_tick_label = []
        for nt in unique_neuron_types:
            y_tick_label.append(nt)
            y_tick.append(np.mean(neuron_order_lookup[np.where([x == nt for x in neuron_type_list])[0]]))

        ax.invert_yaxis()
        ax.set_xlabel('Time (s)', fontsize=12)
        ax.set_yticks(y_tick)
        ax.set_yticklabels(y_tick_label, fontsize=10)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['bottom'].set_linewidth(0.8)

        if skip_time or end_time:
            x_lim = ax.get_xlim()
            x_lim = (0, self.time[1] - skip_time)
            if end_time:
                x_lim = (x_lim[0], end_time)
            ax.set_xlim(x_lim)
        else:
            x_lim = ax.get_xlim()
            x_lim = (self.time[0], self.time[-1])
            ax.set_xlim(x_lim)

        assert not ((skip_time or end_time) and time_range), \
            f"time_range can not be specified with skip_time and end_time"

        if time_range:
            ax.set_xlim(time_range)

        if fig_file is None:
            fig_file = "spike_raster.pdf"

        if title is not None:
            ax.set_title(title)

        fig_file = os.path.join(self.figure_path, fig_file)

        print(f"Saving figure to {fig_file}")
        plt.tight_layout()
        self.make_figures_directory()
        plt.savefig(fig_file, dpi=300)

        plt.ion()
        plt.show()

    def plot_firing_frequency_distribution(self, time_range=None, figure_name=None, bins=20, title=None):

        neuron_types = self.snudda_load.get_neuron_types(return_set=True)

        plt.figure()

        time = self.snudda_simulation_load.get_time()

        for nt in neuron_types:

            neuron_id = self.snudda_load.get_neuron_id_of_type(neuron_type=nt, include_virtual=False)
            spikes = self.snudda_simulation_load.get_spikes(neuron_id=neuron_id)

            if time_range is None:
                freq = [s.size/(time[-1] - time[0]) for s in spikes.values()]
            else:
                freq = [len(np.where(np.logical_and(time_range[0] <= s, s <= time_range[1]))[0])
                        / (time_range[1] - time_range[0])
                        for s in spikes.values()]

            if not isinstance(bins, int):
                if np.max(freq) > max(bins):
                    raise ValueError(f"Max frequency {np.max(freq)} larger than bin range specified {max(bins)}.")

            colour = SnuddaPlotSpikeRaster2.get_colours(nt)
            count, bin = np.histogram(freq, bins=bins)
            plt.stairs(count, bin, label=nt, color=colour, linewidth=3)

        plt.legend()
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Count")

        if title is not None:
            plt.title(title)

        plt.tight_layout()

        if figure_name is not None:
            self.make_figures_directory()

            plt.savefig(os.path.join(self.figure_path, figure_name))

        plt.ion()
        plt.show()

    def plot_population_frequency(self, population_id, time_ranges=None):

        raise NotImplementedError()

        if population_id is None:
            population_id = sorted(list(self.snudda_load.get_neuron_population_units(return_set=True)))
            label = [str(x) for x in population_id]

        if time_ranges is None:
            time_ranges = [(0, np.max(self.snudda_simulation_load.get_time()))]

        x_labels = [f"{x[0]} -- {x[1]}" for x in time_ranges]

        data = dict()

        for pop_id in population_id:
            neuron_id = self.snudda_load.get_population_unit_members(population_unit=pop_id)
            freq_table = self.snudda_simulation_load.get_frequency(neuron_id=neuron_id, time_ranges=time_ranges)
            avg_freq = np.sum(freq_table, axis=0).flatten()

            data[str(pop_id)] = avg_freq

        self.plot_grouped_bars(legend_labels_and_data=data)

    def plot_grouped_bars(self, legend_labels_and_data, x_labels, y_unit_label, title):

        raise NotImplementedError()

        # Derived from matplotlib example

        # x_labels = ("Adelie", "Chinstrap", "Gentoo")
        # legend_labels_and_data = {
        #     'Bill Depth': (18.35, 18.43, 14.98),
        #     'Bill Length': (38.79, 48.83, 47.50),
        #     'Flipper Length': (189.95, 195.82, 217.19),
        # }

        x = np.arange(len(x_labels))  # the label locations
        width = 0.25  # the width of the bars
        multiplier = 0

        fig, ax = plt.subplots(layout='constrained')

        for attribute, measurement in legend_labels_and_data.items():
            offset = width * multiplier
            rects = ax.bar(x + offset, measurement, width, label=attribute)
            ax.bar_label(rects, padding=3)
            multiplier += 1

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel(y_unit_label)
        ax.set_title(title)
        ax.set_xticks(x + width, x_labels)
        ax.legend(loc='upper left', ncols=3)
        ax.set_ylim(0, 250)

        plt.show()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser("Scatter plot")
    parser.add_argument("network_path", type=str, help="Network path")

    args = parser.parse_args()
    ps = SnuddaPlotSpikeRaster2(network_path=args.network_path)

    type_order = ["FS", "dSPN", "LTS", "iSPN", "ChIN"]
    ps.plot_spike_raster(type_order)
    ps.plot_spike_histogram(type_order)



# Visualises neurons and the location of the synapses connecting them.
# Tested using Blender 2.93. Should not be used with older version of Blender.

import bpy
from bpy_extras.io_utils import unpack_list

import os
import mathutils
import numpy as np
from snudda.utils.load import SnuddaLoad
from snudda.utils.snudda_path import snudda_parse_path, get_snudda_data
from snudda.utils.load_network_simulation import SnuddaLoadSimulation


class VisualiseNetwork(object):

    # You need to provide neuron
    def __init__(self, network_path, blender_save_file=None, blender_output_image=None,
                 network_json=None, simulation_output_file_name=None, use_neuron_cache=True):

        self.network_path = network_path
        self.snudda_data = get_snudda_data(network_path=network_path)
        self.scale_f = 1000  # factor to downscale the data
        self.neuron_colour_lookup = dict()  # Allow the user to override the neuron colours
        self.use_neuron_cache = use_neuron_cache

        if network_json:
            self.network_json = network_json
            self.network_file = None
        else:
            self.network_json = None
            self.network_file = os.path.join(network_path, "network-synapses.hdf5")

        if blender_save_file:
            self.blender_save_file = blender_save_file
        else:
            self.blender_save_file = os.path.join(network_path, "visualise-network.blend")

        self.blender_output_image = blender_output_image

        if simulation_output_file_name:
            self.slns = SnuddaLoadSimulation(simulation_output_file_name)
            self.spike_times = self.slns.get_spikes()
        else:
            self.spike_times = None

        self.neuron_cache = dict([])

        # Load the neuron positions
        if self.network_file:
            self.sl = SnuddaLoad(self.network_file)
            self.data = self.sl.data
        elif self.network_json:
            from snudda.utils.fake_load import FakeLoad
            self.sl = FakeLoad()
            self.sl.import_json(self.network_json)
            self.data = self.sl.data

    def set_neuron_colour(self, neuron_id, colour):

        if len(colour) != 4:
            raise ValueError(f"Colour should be R,G,B,alpha (4 values)")
        self.neuron_colour_lookup[neuron_id] = colour

    def clear_neuron_colours(self):

        self.neuron_colour_lookup = dict()

    def set_flat_color(self, mat, rgba):
        '''
        diffuse_color is not the authoritative source of render color in Blender 5.0. The Principled BSDF Base Color is.
        only setting diffuse_color will lead to almost no visible color in renders, so we set both.
        '''
        # Keep diffuse_color for viewport consistency
        mat.diffuse_color = rgba

        # Ensure the render uses the same color
        mat.use_nodes = True
        bsdf = mat.node_tree.nodes.get("Principled BSDF")
        if bsdf is not None:
            bsdf.inputs["Base Color"].default_value = rgba
            bsdf.inputs["Alpha"].default_value = rgba[3]

    def visualise(self,
                  neuron_id=None,
                  blender_output_image=None,
                  white_background=True,
                  show_synapses=True,
                  synapse_colour=None,
                  synapse_pair_filter=None,
                  draw_meshes=True,
                  full_meshes=None,
                  camera_location=None,
                  camera_rotation=None,
                  camera_scale=None,
                  detail_level=1):

        """
            Visualise network in blender.

            Args:
                neuron_id (list) : Neuron ID to visualise, default None means all
                blender_output_image
                white_background
                show_synapses
                synapse_colour: R,G,B,alpha (For values, range 0-1). Default None.
                synapse_pair_filter (list): List of pairs of neurons (tuples) to show synapses for, default None = no filtering
                camera_location
                camera_rotation
                camera_scale
                detail_level (int) : 1 = full morphologies, 2 = reduced morphologies, 3 = soma only

        """

        if neuron_id is not None:
            neurons = [self.data["neurons"][x] for x in neuron_id]
        else:
            neurons = self.data["neurons"]
            neuron_id = self.data["neuron_id"]

        if blender_output_image:
            self.blender_output_image = blender_output_image

        if camera_location is None:
            camera_location = (3.228, 4.3433, 7.6695)  # (2.98, 2.68, 4.96)

        if camera_rotation is None:
            camera_rotation = (94.1*(np.pi/180), 0, -103*(np.pi/180))  # (1.59, 0, -0.26)

        if camera_scale is None:
            camera_scale = (1, 1, 1)

        if synapse_pair_filter is not None:
            synapse_pair_filter = set(synapse_pair_filter)

        origo = self.data["simulation_origo"]
        voxel_size = self.data["voxel_size"]

        # Remove the start cube
        VisualiseNetwork.clean_scene()
        self.neuron_cache.clear() # had the cache error when rendering the whole network so clearing it here

        # Add a light source
        # TODO: Add choice of adding a light source, instead of True here.
        if True:
            sun_location = (10, 10, 10)
            lamp_data = bpy.data.lights.new(name="Sun", type='SUN')

        # Create new object, pass the light data
            sun_object = bpy.data.objects.new(name="sun_object", object_data=lamp_data)

        # Link object to collection in context
            bpy.context.collection.objects.link(sun_object)

        # Change light position
            sun_object.location = sun_location

        bpy.data.scenes['Scene'].render.engine = 'CYCLES'
        world = bpy.data.worlds['World']
        world.use_nodes = True

        # changing these values does affect the render.
        bg = world.node_tree.nodes['Background']
        if white_background:  # Set to True for white background
            bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
            bg.inputs[1].default_value = 1.0
            bpy.context.scene.view_settings.view_transform = 'Standard'

        else:
            bg.inputs[0].default_value[:3] = (0.0, 0.0, 0.0)
            bg.inputs[1].default_value = 0.0

        #Scott's magic
        '''
        mat_dspn = bpy.data.materials.new("DSPN")
        mat_dspn.use_nodes = True
        mat_dspn.node_tree.nodes["Principled BSDF"].inputs[0].default_value = (0, 1, 0, 1)
        mat_dspn.node_tree.nodes["Principled BSDF"].inputs['Alpha'].default_value = 1
        '''
        
        # Define materials
        mat_dspn = bpy.data.materials.new("PKHG")
        # set both diffuse_color and Principled BSDF Base Color
        self.set_flat_color(mat_dspn,(77. / 255, 151. / 255, 1.0, 0.5)) #RGBA
        mat_ispn = bpy.data.materials.new("PKHG")
        self.set_flat_color(mat_ispn,(67. / 255, 55. / 255, 181. / 255, 0.5))
        mat_fs = bpy.data.materials.new("PKHG")
        self.set_flat_color(mat_fs,(6. / 255, 31. / 255, 85. / 255, 1.0))
        mat_chin = bpy.data.materials.new("PKHG")
        self.set_flat_color(mat_chin,(252. / 255, 102. / 255, 0.0, 1.0))
        mat_lts = bpy.data.materials.new("PKHG")
        self.set_flat_color(mat_lts,(150. / 255, 63. / 255, 212. / 255, 1.0))

        mat_snr = bpy.data.materials.new("PKHG")

        # PURPLE

        mat_snr.diffuse_color = (102 / 255, 0 / 255, 102 / 255, 1.0)
        mat_snr.diffuse_color = (102 / 255, 0 / 255, 102 / 255, 1.0)

        # BLUE-PURPLE

        mat_arky = bpy.data.materials.new("PKHG")
        mat_arky.diffuse_color = (43 / 255, 0 / 255, 255 / 255, 1.0)

        # GREEN
        mat_proto = bpy.data.materials.new("PKHG")
        mat_proto.diffuse_color = (0 / 255, 130 / 255, 0 / 255, 1.0)

        mat_other = bpy.data.materials.new("PKHG")
        self.set_flat_color(mat_other,(0.4, 0.4, 0.4, 1.0))
        mat_synapse = bpy.data.materials.new("PKHG")

        material_lookup = {"dspn": mat_dspn,
                           "ispn": mat_ispn,
                           "fsn": mat_fs,
                           "fs": mat_fs,
                           "chin": mat_chin,
                           "lts": mat_lts,
                           "snrneurons": mat_snr,
                           "proto": mat_proto,
                           "arky": mat_arky,
                           "synapse": mat_synapse,
                           "other": mat_other}

        # Add the user requested custom colours
        for nid in self.neuron_colour_lookup.keys():
            material_lookup[nid] = bpy.data.materials.new(str(nid))
            material_lookup[nid].use_nodes = True
            material_lookup[nid].node_tree.nodes["Principled BSDF"].inputs[0].default_value = self.neuron_colour_lookup[nid]
            material_lookup[nid].node_tree.nodes["Principled BSDF"].inputs['Alpha'].default_value = self.neuron_colour_lookup[nid][-1]
                
            #material_lookup[nid] = bpy.data.materials.new("PKHG")
            #material_lookup[nid].diffuse_color = self.neuron_colour_lookup[nid]
        
        if synapse_colour is not None:
            self.set_flat_color(mat_synapse,synapse_colour)
        elif white_background:
            self.set_flat_color(mat_synapse,(0.8, 0.0, 0.0, 1.0))
        else:
            self.set_flat_color(mat_synapse,(1.0, 1.0, 0.9, 1.0))

        # matSynapse.use_transparency = True

        """
        # We comment out these lines, to get the synapse colour to be set correctly (otherwise they are white)
        # Thanks Scott for finding this fix.
        
        mat_synapse.use_nodes = True

        if not white_background:
            emission_strength = 5.0

            # Make synapses glow
            emission = mat_synapse.node_tree.nodes.new('ShaderNodeEmission')
            emission.inputs['Strength'].default_value = emission_strength

            material_output = mat_synapse.node_tree.nodes.get('Material Output')
            mat_synapse.node_tree.links.new(material_output.inputs[0], emission.outputs[0])
        """

        for idx, neuron in enumerate(neurons):

            e_rot = mathutils.Matrix(neuron["rotation"].reshape(3, 3)).to_euler()

            if self.use_neuron_cache and neuron["name"] in self.neuron_cache:
                # If we already have the object in memory, copy it.
                obj = self.neuron_cache[neuron["name"]].copy()

                if self.neuron_cache[neuron["name"]].data:
                    obj.data = self.neuron_cache[neuron["name"]].data.copy()

                VisualiseNetwork.copy_children(self.neuron_cache[neuron["name"]], obj)
                obj.animation_data_clear()

                # Will return None if there is no obj named CUBe
                obj.name = f"{neuron['name']}-{neuron['neuron_id']}"
                VisualiseNetwork.link_object(obj)
            else:
                if type(detail_level) == np.ndarray:
                    if len(detail_level) != len(neurons):
                        raise ValueError(f"detail_level is either 1,2 or 3, "
                                         f"if given as a array must be same length as number of neurons (ie {len(idx)}).")
                    dl = detail_level[idx]
                else:
                    dl = detail_level

                self.read_swc_data(filepath=snudda_parse_path(neuron["morphology"], self.snudda_data), detail_level=dl)
                obj = bpy.context.selected_objects[0]
                obj.name = f"{neuron['name']}-{neuron['neuron_id']}"

                self.neuron_cache[neuron["name"]] = obj

            obj.rotation_euler = e_rot

            print(f"Setting neuron {neuron['neuron_id']} ({neuron['name']}) position: {neuron['position']}")
            obj.location = neuron["position"] * self.scale_f

            n_type = neuron["type"].lower()

            if neuron['neuron_id'] in material_lookup:
                # Custom colour for neuron (priority)
                mat = material_lookup[neuron['neuron_id']]
            elif n_type in material_lookup:
                # Each neuron type has its own colour
                mat = material_lookup[n_type]
            else:
                mat = material_lookup["other"]

            if self.spike_times:
                rest_color = mat.diffuse_color[:]
                # if animating spike times we need to make a fresh material per neuron
                mat_spikes = bpy.data.materials.new("PKHG")
                mat_spikes.diffuse_color = rest_color
                mat_spikes.keyframe_insert(data_path="diffuse_color", frame=1.0, index=-1)
                if str(neuron['neuron_id']) in self.spike_times.keys():
                    spikes = self.spike_times[str(neuron['neuron_id'])]

                    # convert 'time' to Blender frames; factor of ~100 works nicely
                    spike_frames = np.round(100 * np.array(spikes))
                    for t in spike_frames:
                        mat_spikes.diffuse_color = rest_color
                        # need to add an instruction to remain at rest colour at a pre-spike time
                        # so that the colour change is not gradual but quasi-instantaneous
                        mat_spikes.keyframe_insert(data_path="diffuse_color", frame=t - 1, index=-1)
                        mat_spikes.diffuse_color = (1, 1, 1, 1)
                        mat_spikes.keyframe_insert(data_path="diffuse_color", frame=t, index=-1)
                        mat_spikes.diffuse_color = rest_color
                        # change back to rest colour
                        mat_spikes.keyframe_insert(data_path="diffuse_color", frame=t + 5, index=-1)
                print("Color......")
                for ch in obj.children:
                    ch.active_material = mat_spikes
            else:
                print("Color......")
                for ch in obj.children:
                    ch.active_material = mat

            obj.select_set(False)

        if show_synapses:
            print("Adding synapses...")
            self.syn_coll = bpy.data.collections.new(name="Synapses") 
            bpy.context.scene.collection.children.link(self.syn_coll)
            # Draw the synapses
            n_synapses = 0

            for ob in bpy.context.selected_objects:
                ob.select_set(False)
            synapse_obj = None

            for vis_pre_id in neuron_id:
                for vis_post_id in neuron_id:
                    synapses, synapse_coords = self.sl.find_synapses(pre_id=vis_pre_id, post_id=vis_post_id)

                    if synapses is None:
                        # No synapses between pair
                        continue

                    for syn in synapses:
                        pre_id = syn[0]
                        post_id = syn[1]

                        if synapse_pair_filter is not None:
                            if (pre_id, post_id) not in synapse_pair_filter:
                                continue

                        assert pre_id == vis_pre_id and post_id == vis_post_id  # Just sanity check, should be true

                        # Draw this neuron (the SWC import scales from micrometers to mm), the
                        # positions in the simulation are in meters, need to scale it to mm for
                        # blender to have same units.
                        x = (origo[0] + voxel_size * syn[2]) * self.scale_f
                        y = (origo[1] + voxel_size * syn[3]) * self.scale_f
                        z = (origo[2] + voxel_size * syn[4]) * self.scale_f

                        if synapse_obj:
                            obj = synapse_obj.copy()
                            if synapse_obj.data:
                                obj.data = synapse_obj.data.copy()
                            obj.animation_data_clear()
                            obj.location = (x, y, z)
                            obj.name = f"synapse-{n_synapses}"
                            VisualiseNetwork.link_object(obj)

                        else:
                            bpy.ops.mesh.primitive_uv_sphere_add(radius=0.001 * 4, location=(x, y, z), scale=(1, 1, 1),
                                                                 segments=16, ring_count=8)
                            obj = bpy.context.selected_objects[0]
                            obj.active_material = mat_synapse
                            obj.select_set(False)
                            synapse_obj = obj

                        n_synapses += 1

                        # print(f"Added synapse #{n_synapses} at {[x, y, z]}")
                        if n_synapses % 5000 == 0:
                            print(f"Synapses added so far: {n_synapses}")
            bpy.ops.object.select_all( action='DESELECT' )
            bpy.ops.object.select_pattern(pattern="synapse*")
            s_objs = bpy.context.selected_objects
            for s in s_objs:
                self.syn_coll.objects.link(s)
            print(f"nSynapses = {n_synapses}")

        if draw_meshes:
            self.struct_coll = bpy.data.collections.new(name="Structures") #create new coll in data
            bpy.context.scene.collection.children.link(self.struct_coll) #add new coll to the scene
            self.add_all_meshes()

        if full_meshes:
            for struct, mesh_file in full_meshes.items():
                self.add_mesh_structure(mesh_file=snudda_parse_path(mesh_file, self.snudda_data), colour=(0.1, 0.1, 0.1),
                                    alpha=0.1)

        bpy.ops.object.camera_add(enter_editmode=False, align='VIEW',
                                  location=camera_location,
                                  rotation=camera_rotation,
                                  scale=camera_scale)
        cam = bpy.data.objects["Camera"]
        print(bpy.context.selected_objects)
        bpy.context.scene.camera = cam

        bpy.ops.wm.save_as_mainfile(filepath=self.blender_save_file)

        if self.blender_output_image:
            # When testing (e.g. adjusting camera position) you can skip this and just look at the blender file directly
            print("Rendering image.")
            bpy.ops.render.render()
            bpy.data.images['Render Result'].save_render(filepath=self.blender_output_image)

    def add_mesh_structure(self, mesh_file, colour, alpha):

        mat = bpy.data.materials.new("PKHG")
        mat.diffuse_color = (colour[0], colour[1], colour[2], alpha)
        mat.use_nodes = True
        mat.node_tree.nodes["Principled BSDF"].inputs['Alpha'].default_value = alpha
        mat.node_tree.nodes["Principled BSDF"].inputs['Base Color'].default_value = (colour[0], colour[1], colour[2], alpha)

        # --- Import OBJ (Blender 5.0) ---
        bpy.ops.wm.obj_import(
            filepath=mesh_file,
            forward_axis='Y',
            up_axis='Z'
        )

        o = bpy.context.selected_objects[0]
        # scale_f = 1000
        o.scale[0] = 1 / self.scale_f
        o.scale[1] = 1 / self.scale_f
        o.scale[2] = 1 / self.scale_f
        o.active_material = mat
        self.struct_coll.objects.link(o)
    def add_all_meshes(self):
        # Iterate over all regions in the config
        for region_name, region_cfg in self.sl.config["regions"].items():

            # Each region is expected to have a "volume" block
            vol_cfg = region_cfg.get("volume")

            mesh_path = vol_cfg.get("mesh_file")

            self.add_mesh_structure(
                mesh_file=snudda_parse_path(mesh_path, self.snudda_data),
                colour=(0.1, 0.1, 0.1),
                alpha=0.1
            )

    @staticmethod
    def copy_children(parent, parent_copy):
        for child in parent.children:
            child_copy = child.copy()
            child_copy.parent = parent_copy
            VisualiseNetwork.link_object(child_copy)
            VisualiseNetwork.copy_children(child, child_copy)

    @staticmethod
    def link_object(obj):
        try:
            bpy.context.collection.objects.link(obj)
        except:
            print("Blender 2.8 failed. Likely due to 2.7 syntax.")
            # bpy.context.scene.objects.link(obj)  # Blender 2.7

    @staticmethod
    def clean_scene():
        # add robust scene cleaning
        print("Cleaning the scene.")

        scene = bpy.context.scene

        # Remove all objects linked to the current scene
        for obj in list(scene.objects):
            bpy.data.objects.remove(obj, do_unlink=True)

        for mesh in list(bpy.data.meshes):
            if mesh.users == 0:
                bpy.data.meshes.remove(mesh)

        for mat in list(bpy.data.materials):
            if mat.users == 0:
                bpy.data.materials.remove(mat)


    def read_swc_data(self, filepath, detail_level=1):

        """
            Read SWC file

            Args:
                filepath (str) : Path to SWC file
                detail_level (int) : Detail level 1 = full detail, 2 = reduced quality, 3 = soma only

        """

        ''' read swc file '''
        print(filepath)
        f = open(filepath)
        lines = f.readlines()
        f.close()

        ''' find starting point '''
        x = 0
        while lines[x][0] == '#':
            x += 1

        ''' Create a dictionary with the first item '''
        data = lines[x].strip().split(' ')
        soma_id = int(data[0])
        soma_type = float(data[1])
        soma_x = float(data[2])
        soma_y = float(data[3])
        soma_z = float(data[4])
        soma_r = float(data[5])
        soma_parent = int(data[6])

        # We centre the neuron
        neuron = {soma_id: [soma_type, 0.0, 0.0, 0.0, soma_r, soma_parent]}
        x += 1

        ''' Read the rest of the lines to the dictionary '''
        for ls in lines[x:]:
            data = ls.strip().split(' ')
            comp_id = int(data[0])
            comp_type = float(data[1])
            comp_x = float(data[2])
            comp_y = float(data[3])
            comp_z = float(data[4])
            comp_r = float(data[5])
            comp_parent = int(data[6])

            # Centre neuron, so soma is at 0,0,0
            neuron[comp_id] = [comp_type, comp_x - soma_x, comp_y - soma_y, comp_z - soma_z, comp_r, comp_parent]

        bpy.ops.object.empty_add(type='ARROWS',
                                 location=(
                                     neuron[1][1] / self.scale_f, neuron[1][2] / self.scale_f,
                                     neuron[1][3] / self.scale_f),
                                 rotation=(0, 0, 0))
        a = bpy.context.selected_objects[0]
        a.name = 'neuron_swc'

        last = -10.0

        line_points = []
        line_radius = []

        ''' Create object '''
        for key, value in neuron.items():
            # value contains: 0: type, 1: x, 2: y, 3: z, 4: r, 5: parent

            if value[0] == 1:

                if detail_level > 1:
                    segments = 10
                    ring_count = 5
                else:
                    segments = 32
                    ring_count = 16

                # This is the soma, add it
                soma_radie = value[-2]
                bpy.ops.mesh.primitive_uv_sphere_add(segments=segments,
                                                     ring_count=ring_count,
                                                     location=(
                                                         value[1] / self.scale_f, value[2] / self.scale_f,
                                                         value[3] / self.scale_f),
                                                     radius=soma_radie / self.scale_f)
                soma_obj = bpy.context.selected_objects[0]
                soma_obj.parent = a

                print(f"Adding soma {value}")

            if value[-1] == -1:
                continue

            if value[0] == 10:
                continue

            if value[-1] == last:
                line_points.append(
                    [neuron[value[-1]][1] / self.scale_f, neuron[value[-1]][2] / self.scale_f,
                     neuron[value[-1]][3] / self.scale_f])
                line_radius.append(neuron[value[-1]][4] / self.scale_f)
            else:
                # Add Bezier curve for previous data
                VisualiseNetwork.add_bezier(curve_parent=a,
                                            line_points=line_points,
                                            line_radius=line_radius,
                                            detail_level=detail_level)
                line_points = []
                line_radius = []

            last = key

        # Add the last line
        VisualiseNetwork.add_bezier(curve_parent=a,
                                    line_points=line_points,
                                    line_radius=line_radius,
                                    detail_level=detail_level)
        line_points = []
        line_radius = []

        a.select_set(True)

        return {'FINISHED'}

    @staticmethod
    def add_bezier(curve_parent, line_points, line_radius, detail_level=1):

        if len(line_points) == 0:
            return

        if detail_level > 2:
            # Only soma
            return

        if detail_level == 2:
            # Keep only end points and middle point
            if len(line_radius) > 3:
                line_points = [line_points[0],
                               line_points[int(len(line_radius) / 2)],
                               line_points[-1]]
                line_radius = [line_radius[0],
                               line_radius[int(len(line_radius) / 2)],
                               line_radius[-1]]

        tracer = bpy.data.curves.new('tracer', 'CURVE')
        tracer.dimensions = '3D'
        spline = tracer.splines.new('BEZIER')
        spline.bezier_points.add(len(line_points) - 1)

        curve = bpy.data.objects.new('curve', tracer)
        curve.data.use_fill_caps = True  # Added 2019-06-17
        curve.parent = curve_parent

        bpy.context.scene.collection.objects.link(curve)

        # render ready curve
        tracer.resolution_u = 8
        tracer.bevel_resolution = 8
        tracer.fill_mode = 'FULL'
        tracer.bevel_depth = 1.0

        # move nodes to objects
        # spline.bezier_points.foreach_set("co", unpack_list(line_points))
        # spline.bezier_points.foreach_set("radius", unpack_list(line_radius))
        # spline.bezier_points.foreach_set("handle_right_type", unpack_list(["VECTOR"] * len(line_radius)))
        # spline.bezier_points.foreach_set("handle_left_type", unpack_list(["VECTOR"] * len(line_radius)))

        for p, r, co in zip(spline.bezier_points, line_radius, line_points):
            p.radius = r
            p.co = co
            p.handle_right_type = "VECTOR"
            p.handle_left_type = "VECTOR"

# TODO: Look for speedup -- https://blender.stackexchange.com/questions/7358/python-performance-with-blender-operators
# TODO: https://blenderartists.org/t/python-slowing-down-over-time/569534/8



from snudda.init.init import SnuddaInit
from collections import OrderedDict
import json
import os

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Init custom network")
    parser.add_argument("network", type=str, help="Network path")
    parser.add_argument("--neurons", "--cellspec", type=str, help="Neurons directory", default=None)
    # TODO: Add cell number parameters
    args = parser.parse_args()

    sim_name = args.network
    neurons_dir = args.neurons

    connect_neurons = True

    # simName = "networks/FSmorphTest2orig"
    # simName = "networks/FSmorphTest1b"
    # simName = "LTStest"
    # simName = "networks/twoFS"
    # simName = "networks/FSmorphTest4"
    # simName = "networks/3types-striatum"
    # simName = "networks/SynTest-v6" # MSMS tuning
    # simName = "networks/SynTest-v15"

    config_name = os.path.join(sim_name, "network-config.json")
    cnc = SnuddaInit(struct_def={}, config_file=config_name)
    # cnc.defineStriatum(nMSD1=500,nMSD2=500,nFS=0,nLTS=0,nChIN=30,volumeType="cube")
    # cnc.defineStriatum(nMSD1=120,nMSD2=120,nFS=20,nLTS=0,nChIN=0,volumeType="slice")
    # cnc.defineStriatum(nMSD1=0,nMSD2=0,nFS=10000,nLTS=0,nChIN=0,volumeType="slice")
    # cnc.defineStriatum(nMSD1=0,nMSD2=0,nFS=10000,nLTS=0,nChIN=0,volumeType="cube")
    # cnc.define_striatum(num_dSPN=0, num_iSPN=0, num_FS=100, num_LTS=0, num_ChIN=0, volume_type="cube")
    # cnc.defineStriatum(nMSD1=10,nMSD2=10,nFS=10,nLTS=10,nChIN=10,volumeType="slice")

    # cnc.defineStriatum(nMSD1=500,nMSD2=500,nFS=0,nLTS=0,nChIN=500,volumeType="cube")
#    cnc.define_striatum(num_dSPN=1500, num_iSPN=1500, num_FS=0, num_LTS=0, num_ChIN=0,
#                        volume_type="cube", neurons_dir=neurons_dir)
    # cnc.define_striatum(num_dSPN=47500, num_iSPN=47500, num_FS=1300, num_LTS=0, num_ChIN=0,
    #                     volume_type="cube", neurons_dir=neurons_dir)
    cnc.define_striatum(num_dSPN=25000, num_iSPN=25000, num_FS=500, num_LTS=0, num_ChIN=0,
                        volume_type="cube", neurons_dir=neurons_dir)
    cnc.write_json(config_name)

    if not connect_neurons:
        print("Removing all target information, and rewriting config file")
        # Reopen the config file, and remove all connectivity settings, to
        # get an unconnected network

        with open(config_name, "r") as f:
            con_data = json.load(f, object_pairs_hook=OrderedDict)

        for k in con_data:
            if k.lower() in ["volume", "channels"]:
                continue

            # Remove targets
            if False:
                x = con_data[k]
                del x["targets"]

        with open(config_name, "w") as f:
            print("Writing to file: " + str(config_name))
            json.dump(con_data, f, indent=2)

    print("Now run:\nsnudda place " + sim_name)
    print("snudda detect " + sim_name)
    print("snudda prune " + sim_name)
    print("snudda input " + sim_name + " --input config/input-tinytest-v4.json")
    print("snudda simulate " + sim_name \
          + " --voltOut " + sim_name + "/volt-out.csv --time 10.0")



import unittest
import os
from snudda.utils.snudda_path import snudda_parse_path, get_snudda_data, snudda_simplify_path


class TestSnuddaPath(unittest.TestCase):

    def test_parse(self):

        self.assertEqual(snudda_parse_path("$DATA/test", "/abc"), "/abc/test")

        self.assertEqual(snudda_parse_path("$SNUDDA_DATA/test", "abc/def/"),
                         os.path.abspath(os.path.join(os.path.dirname(__file__), "abc/def/test")))

    def test_get(self):

        self.assertEqual(get_snudda_data(),
                         os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "snudda", "data")))

        self.assertEqual(get_snudda_data(snudda_data=os.path.dirname(__file__)),
                         os.path.dirname(__file__))

        # A few more cases to test...

    def test_simplify(self):

        self.assertEqual(snudda_simplify_path("/abc/def", "/abc"), "$SNUDDA_DATA/def")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.place.region_mesh_vedo import RegionMesh


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:

        mesh_path = "my_test_mesh.obj"
        centre = (100e-6, 110e-6, 120e-6)
        side_len = 100e-6
        create_cube_mesh(mesh_path, centre, side_len)
        self.rm = RegionMesh(mesh_path=mesh_path)

    def test_inside(self):

        rng = np.random.default_rng(1)
        test_points = rng.uniform(low=-100e-6, high=300e-6, size=(1000, 3))

        inside_flag = self.rm.check_inside(test_points)

        ref_inside_flag = (50e-6 <= test_points[:, 0]) & (test_points[:, 0] <= 150e-6) \
            & (60e-6 <= test_points[:, 1]) & (test_points[:, 1] <= 160e-6) \
            & (70e-6 <= test_points[:, 2]) & (test_points[:, 2] <= 170e-6)

        self.assertTrue((inside_flag == ref_inside_flag).all())

    def test_distance(self):

        test_points = np.array([[40e-6, 110e-6, 120e-6],  # distance 10e-6
                                [100e-6, 90e-6, 120e-6],  # distane -20e-6
                                [160e-6, 170e-6, 180e-6],  # distance np.sqrt(3*100)
                                ])

        ref_distance = np.array([10e-6, -30e-6, np.sqrt(3*100e-12)])

        dist = self.rm.distance_to_border(test_points)

        self.assertTrue(np.allclose(dist, ref_distance, rtol=1e-7))


if __name__ == '__main__':
    unittest.main()



import neuron
import numpy as np
import sys


# Plot all sections
# [neuron.h.psection(x) for x in neuron.h.allsec()]

##############################################################################

class RunLittleSynapseRun(object):

    ############################################################################

    def __init__(self,
                 stim_times,
                 holding_voltage=-70e-3,
                 synapse_type='glut',
                 params={},
                 time=2.0):

        # This just clears the sections, just to be on the safe side
        # !!! Didnt work... still problem
        # for s in neuron.h.allsec():
        #  import pdb
        #  pdb.set_trace()
        #  neuron.h.delete_section(sec=s)

        print(f"Holding voltage: {holding_voltage} V")
        print(f"Stim times: {stim_times} s")
        print(f"Synapse type: {synapse_type}")

        self.time = time

        neuron.h.load_file('stdrun.hoc')

        # Should we use weak reference for garbage collection? (weakref package)

        self.soma = neuron.h.Section(name='soma')
        self.soma.insert('hh')
        neuron.h.celsius = 35

        # gnabar_hh: The maximum specific sodium channel conductance [Default value = 0.120 S/cm2]
        # gkbar_hh: The maximum specific potassium channel conductance [Default value = 0.036 S/cm2]
        # gl_hh: The maximum specific leakage conductance [Default value = 0.0003 S/cm2]
        # ena: The reversal potential for the sodium channel [Default value = 50 mV]
        # ek: The reversal potential for the potassium channel [Default value = -77 mV]
        # el_hh: The reversal potential for the leakage channel [Default value = -54.3 mV]

        try:
            if synapse_type.lower() == 'glut':
                self.little_synapse = neuron.h.tmGlut(self.soma(0.5))
            elif synapse_type.lower() == "gaba":
                self.little_synapse = neuron.h.tmGabaA(self.soma(0.5))
            else:
                assert False, f"Unknown synapse type: {synapse_type}"

        except:
            import traceback
            tstr = traceback.format_exc()
            print(tstr)

            print("Did you remember to run nrnivmodl first, to generate channels mod files?")
            sys.exit(-1)

        # We need to set the params also
        self.params = params

        self.default_cond = 5e-7

        for p in self.params:

            # Conductance is set separately, it is a property of netcon
            if p == "cond":
                self.default_cond = self.params["cond"]
                continue

            if p == "somaDiameter":
                print(f"Setting soma diameter to {self.params['somaDiameter'] * 1e6} mum")
                self.soma.L = self.params["somaDiameter"] * 1e6
                self.soma.diam = self.params["somaDiameter"] * 1e6
                continue

            if p == "somaCM":
                print(f"Setting soma CM to {self.params['somaCM'] * 1e2} uf/cm2")
                self.soma.cm = self.params["somaCM"] * 1e2
                continue

            if p == "somaGleak":
                print("Setting GL_HH in soma to {self.params['somaGleak'] * 1e-4} S/cm2")
                self.soma.gl_hh = self.params["somaGleak"] * 1e4
                continue

            val = self.si_to_natural_units(p, self.params[p])

            setattr(self.little_synapse, p, val)
            print(f"Setting parameters: {p} = {val} (neuron natural units)")

        self.vec_stim = neuron.h.VecStim()

        self.stim_times = stim_times * 1e3

        self.stim_vector = neuron.h.Vector(stim_times * 1e3)
        self.vec_stim.play(self.stim_vector)

        self.nc_syn = neuron.h.NetCon(self.vec_stim, self.little_synapse)
        self.nc_syn.delay = 0
        self.nc_syn.threshold = 0

        self.t_save = neuron.h.Vector()
        self.t_save.record(neuron.h._ref_t)

        self.v_save = neuron.h.Vector()
        self.v_save.record(self.soma(0.5)._ref_v)

        # Channel current save
        self.i_save = neuron.h.Vector()
        self.i_save.record(self.little_synapse._ref_i)

        self.holding_voltage = holding_voltage

        self.v_clamp = neuron.h.SEClamp(self.soma(0.5))

        self.v_clamp_rs = 1e-9  # 1e-6
        self.v_clamp.rs = self.v_clamp_rs
        self.v_clamp.amp1 = holding_voltage * 1e3
        self.v_clamp.dur1 = self.time * 2 * 1e3
        # print("VClamp duration: " + str(self.VClamp.dur1))

        neuron.h.finitialize(self.holding_voltage * 1e3)
        # !!! There is a WEIRD neuron bug, that if this tstop here is
        # different from duration of simulation, then the *SECOND* time
        # a model is initialised we get the length of tSave set by this
        # value, and not by the tStop of that simulation --- go figure!
        neuron.h.tstop = self.time * 1e3  # Must set tstop
        neuron.h.run()

        # neuron.h.tstop = 2001
        # neuron.h.finitialize(self.holdingVoltage*1e3)
        # neuron.h.run()

        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(self.t_save, self.v_save)
            plt.title("Holding voltage should be " \
                      + str(self.holding_voltage * 1e3) + "mV")
            plt.xlabel("time (ms)")
            plt.ylabel("volt (mV)")
            plt.ion()
            plt.show()
            import pdb
            pdb.set_trace()

            import pdb
            pdb.set_trace()

        cur = float(self.v_clamp.i)

        # Remove VClamp
        self.v_clamp = None

        self.i_clamp = neuron.h.IClamp(self.soma(0.5))
        self.i_clamp.amp = cur  # nA
        self.i_clamp.dur = 2 * self.time * 1e3

        print(f"Holding voltage {holding_voltage * 1e3} mV, IClamp amp = {cur} nA")

    ############################################################################

    def __del__(self):

        # This should not be needed but...
        self.soma = None
        self.i_clamp = None
        self.v_clamp = None
        self.v_save = None
        self.t_save = None
        self.i_save = None
        self.nc_syn = None

        self.vec_stim = None
        self.stim_vector = None
        self.little_synapse = None

    ############################################################################

    def update_holding_current(self, holding_voltage=None):

        if holding_voltage is None:
            holding_voltage = self.holding_voltage
        else:
            self.holding_voltage = holding_voltage

        # Disable old iClamp temporarily
        self.i_clamp.amp = 0

        # Setup a temporary VClamp
        self.v_clamp = neuron.h.SEClamp(self.soma(0.5))
        self.v_clamp.rs = self.v_clamp_rs
        self.v_clamp.amp1 = holding_voltage * 1e3
        self.v_clamp.dur1 = self.time * 2 * 1e3
        # print("VClamp duration: " + str(self.VClamp.dur1))

        neuron.h.finitialize(self.holding_voltage * 1e3)
        # !!! There is a WEIRD neuron bug, that if this tstop here is
        # different from duration of simulation, then the *SECOND* time
        # a model is initialised we get the length of tSave set by this
        # value, and not by the tStop of that simulation --- go figure!
        neuron.h.tstop = self.time * 1e3  # Must set tstop
        neuron.h.run()

        cur = float(self.v_clamp.i)

        # Remove VClamp
        self.v_clamp = None

        # Update current on iClamp
        self.i_clamp.amp = cur
        self.i_clamp.dur = 2 * self.time * 1e3

    ############################################################################

    def set_stim_times(self, stim_times):

        if len(stim_times) != len(self.stim_times) or (stim_times != self.stim_times).any():
            print(f"Setting stim times to {stim_times} s")
            self.stim_vector = neuron.h.Vector(stim_times * 1e3)
            self.stim_times = stim_times * 1e3

    ############################################################################

    def run(self, tau, tau_r, tau_f, u, cond=None, time=None):

        if time is None:
            time = self.time
        else:
            self.time = time

        if cond is None:
            cond = self.default_cond

        # print(vars())

        # print("Running: tau: %.3g, tauR: %.3g, tauF: %.3g, U: %.3g, cond: %.3g\n" \
        #      % (tau,tauR,tauF,U,cond))

        # Convert from SI units to natural units that Neuron uses
        self.nc_syn.weight[0] = 1 * cond * 1e6
        self.little_synapse.tau = tau * 1e3
        self.little_synapse.tauR = tau_r * 1e3
        self.little_synapse.tauF = tau_f * 1e3
        self.little_synapse.U = u

        # print(self.littleSynapse.tau)

        # print("Initialise voltage to " + str(self.holdingVoltage*1e3) + " mV")
        neuron.h.finitialize(self.holding_voltage * 1e3)  # OLD : -70
        neuron.h.tstop = time * 1e3

        neuron.h.run()

        # self.tSave.resize()
        # self.vSave.resize()
        # self.iSave.resize()

        if np.array(self.t_save).shape != np.array(self.v_save).shape:
            print("THIS IS WRONG, should be same shape!!")
            print(f"size t = {np.array(self.t_save).shape}")
            print(f"size v = {np.array(self.v_save).shape}")
            import pdb
            pdb.set_trace()

        # print("Last V = " + str(self.vSave[-1]*1e-3))

        # Convert back to SI units
        return (np.array(self.t_save) * 1e-3,
                np.array(self.v_save) * 1e-3,
                np.array(self.i_save) * 1e-9)

    ############################################################################

    # I wish Neuron would use natural units...

    def si_to_natural_units(self, var_name, value):

        conv_factor = {"U": 1.0,
                       "tauR": 1e3,
                       "tauF": 1e3,
                       "cond": 1e6,
                       "tau": 1e3,
                       "nmda_ratio": 1.0}

        if var_name not in conv_factor:
            print(f"Missing conversion fractor for {var_name}. Please update SItoNaturalUnits function.")
            print("convFactor = " + str(conv_factor))
            import pdb
            pdb.set_trace()

        try:
            return value * conv_factor[var_name]
        except:
            import traceback
            tstr = traceback.format_exc()
            print(tstr)
            import pdb
            pdb.set_trace()

    ############################################################################

    # pars = { "tau1" : 0.25e-3 }
    # The code converts to natural units internally, if your parameter is missing
    # then update SItoNaturalUnits

    # OBS, soma parameters are ignored by run2 (they should be set during setup)

    def run2(self, pars, time=None, cond=1e-8):

        if time is None:
            time = self.time
        else:
            self.time = time

        for p in pars:
            if p in ["somaDiameter", "somaCM", "somaGleak"]:
                # Ignore soma parameters at this stage, should be set at setup
                continue

            if p == "cond":
                cond = self.si_to_natural_units(p, pars[p])
            else:
                v = self.si_to_natural_units(p, pars[p])
                setattr(self.little_synapse, p, v)

        neuron.h.finitialize(self.holding_voltage * 1e3)
        self.nc_syn.weight[0] = cond
        # print("Synapse conductance: " + str(cond) + " uS")
        # print("Verification of conductance: " + str(self.ncSyn.weight[0]))

        neuron.h.tstop = time * 1e3
        neuron.h.run()

        # Convert results back to SI units
        return (np.array(self.t_save) * 1e-3,
                np.array(self.v_save) * 1e-3,
                np.array(self.i_save) * 1e-9)


##############################################################################

if __name__ == "__main__":

    import matplotlib.pyplot as plt
    import time

    start = time.time()

    stim_times = np.array([0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 1.0])
    tau = 10e-3
    tau_r = 200e-3
    tau_f = 900e-3
    u = 0.5

    tau = 0.1
    tau_r = 0.1
    tau_f = 0.1
    u = 0.1
    cond = 1e-7

    rlsr = RunLittleSynapseRun(stim_times, synapse_type="glut")

    for i in range(10):
        t_s, v_s, i_s = rlsr.run(tau * i, tau_r, tau_f, u, cond)
        plt.figure()
        plt.plot(t_s, v_s)
        plt.xlabel("Time (s)")
        plt.ylabel("Voltage (V)")
        plt.figure()
        plt.plot(t_s, i_s)
        plt.xlabel("Time (s)")
        plt.ylabel("Current (A)")

    # tS,vS = rlsr.run(tau*1,tauR,tauF,U,cond)
    # plt.plot(tS,vS)

    plt.xlabel("Time")
    plt.ylabel("Voltage")
    plt.title("How tau affects dynamics")

    end = time.time()
    print(end - start)

    plt.ion()
    plt.show()

    import pdb

    pdb.set_trace()



#!/usr/bin/env python3

# This is useful if the user wants to purge a particular morphology variation
# from the meta.json file.

import json
import argparse
from pathlib import Path

def filter_morphologies(input_file, output_file, filter_string):
    """Remove morphology keys where morphology contains the filter string."""
    with open(input_file, 'r') as f:
        data = json.load(f)
    
    # Filter out morphology keys with matching morphologies
    for param_key in list(data.keys()):
        for morph_key in list(data[param_key].keys()):
            morphology = data[param_key][morph_key].get('morphology', '')
            if filter_string in morphology:
                del data[param_key][morph_key]
        
        # Remove parameter key if empty after filtering
        if not data[param_key]:
            del data[param_key]
    
    # Write filtered data
    with open(output_file, 'w') as f:
        json.dump(data, f, indent=4)
    
    print(f"Filtered data written to {output_file}")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Remove morphology variations from meta.json file'
    )
    parser.add_argument('input', help='Input meta.json file')
    parser.add_argument('output', help='Output meta.json file')
    parser.add_argument('filter', help='String to filter (e.g., "var1")')
    
    args = parser.parse_args()
    filter_morphologies(args.input, args.output, args.filter)
